""" Tests for spatial occupancy"""
from opexebo.general import shuffle as func
from opexebo import errors

import numpy as np
import pytest

print("=== tests_general_shuffle ===")

###############################################################################
################                MAIN TESTS
###############################################################################


t_start = 1
t_stop = 13
offset_lim = 2.3
iterations = 1000
times = np.arange(t_start + 0.5, t_stop - 0.5)


def test_increment_creation():
    """
    Verify that the increments are correctly created between (t_start + offset_lim) and (t_stop - offset_lim)
    """

    out, inc = func(times, offset_lim, iterations, t_start, t_stop)
    assert inc.size == iterations
    assert min(inc) >= t_start + offset_lim
    assert (
        min(inc) <= t_start + offset_lim + 0.5
    )  # this is a random test, so it _could_ fail...
    assert max(inc) <= t_stop - offset_lim
    assert max(inc) >= t_stop - offset_lim - 0.5


def test_output_creation():
    """
    verify that the output array is the correct shape and size, can be indexed correctly
    """

    out, inc = func(times, offset_lim, iterations, t_start, t_stop)
    assert out.shape == (iterations, times.size)
    assert out[0].shape == times.shape


def test_output_logic():
    """
    Verify that the output is a shuffled copy of the input
    """

    out, inc = func(times, offset_lim, iterations, t_start, t_stop)

    assert np.min(out) >= t_start
    assert np.max(out) <= t_stop
    for row in out:
        assert np.array_equal(row, np.array(sorted(row)))


def test_edge_cases():
    """
    Test some obvious edge cases
    """
    # Basic arguments
    t_start = 1
    t_stop = 13
    offset_lim = 2.3
    iterations = 1000

    times = np.arange(t_start + 0.5, t_stop - 0.5)

    # Few spikes compared to large time range
    out, _ = func(times, 1e3, iterations, t_start, 1e4)
    assert np.array_equal(np.diff(times), np.diff(out[0]))

    # One spike
    spk1 = np.array([5.3])
    out, _ = func(spk1, offset_lim, iterations, t_start, t_stop)
    assert out.shape == (iterations, spk1.size)


"""
Verify that the defensive logic covers as many stupid arguments as I can think of
"""


invalid_times = [
    np.ones((2, 2)),  # 2d data
    3,  # non-array data
    np.full(4, np.nan),  # non-finite data
    "a",
]
invalid_offset = [
    -1,  # negative
    0,
    8,  # too high, i.e. greater than half the difference
    np.nan,  # nonfinite
]
invalid_iterations = [
    -1,  # negative
    0,
]
invalid_t_start = [
    np.nan,
    np.inf,
    1.6,  # larger than min(times)
]
invalid_t_stop = [
    np.nan,
    np.inf,
    10,  # smaller than max(times)
    -1,  # smaller than t_start
]


@pytest.mark.parametrize("invalid_data", invalid_times, ids=str)
def test_invalid_times(invalid_data):
    with pytest.raises(errors.ArgumentError):
        func(invalid_data, offset_lim, iterations, t_start, t_stop)


@pytest.mark.parametrize("invalid_data", invalid_offset, ids=str)
def test_invalid_offsets(invalid_data):
    with pytest.raises(errors.ArgumentError):
        func(times, invalid_data, iterations, t_start, t_stop)


@pytest.mark.parametrize("invalid_data", invalid_iterations, ids=str)
def test_invalid_iterations(invalid_data):
    with pytest.raises(errors.ArgumentError):
        func(times, offset_lim, invalid_data, t_start, t_stop)


@pytest.mark.parametrize("invalid_data", invalid_t_start, ids=str)
def test_invalid_t_start(invalid_data):
    with pytest.raises(errors.ArgumentError):
        func(times, offset_lim, iterations, invalid_data, t_stop)


@pytest.mark.parametrize("invalid_data", invalid_t_stop, ids=str)
def test_invalid_t_stop(invalid_data):
    with pytest.raises(errors.ArgumentError):
        func(times, offset_lim, iterations, t_start, invalid_data)



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  4 11:37:57 2019

@author: simoba
"""

import numpy as np


def fit_ellipse(X, Y):
    '''
    Fit an ellipse to the provided set of X, Y co-ordinates

    Based on the approach taken in
    Authors: Andrew Fitzgibbon, Maurizio Pilu, Bob Fisher
    Reference: "Direct Least Squares Fitting of Ellipses", IEEE T-PAMI, 1999
        
    @Article{Fitzgibbon99,
    author = "Fitzgibbon, A.~W.and Pilu, M. and Fisher, R.~B.",
    title = "Direct least-squares fitting of ellipses",
    journal = pami,
    year = 1999,
    volume = 21,
    number = 5,
    month = may,
    pages = "476--480"
    }
    and implemented in MatLab by Vadim Frolov

    Parameters
    ----------
    X - np.ndarray
        x co-ordinates of points
    Y - np.ndarray
        y co-ordinates of points

    Returns
    -------
    x_centre : float
        Centre of ellipse
    y_centre : float
        Centre of ellipse
    Ru : float
        Major radius
    Rv : float
        Minor radius
    theta_rad : float
        Ellipse orientation (in radians)


    Notes
    --------
    BNT.+general.fitEllipse
    opexebo.analysis.gridscore

    Copyright (C) 2019 by Simon Ball

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    '''

    if X.size != Y.size:
        raise ValueError("X and Y must be the same length. You provided"\
                         " (%d, %d) values respectively" % (X.size, Y.size))
    if not np.isfinite(X).all():
        raise ValueError("X cannot contain values that are nan or inf."\
                         f" You provided {X}")
    if not np.isfinite(Y).all():
        raise ValueError("X cannot contain values that are nan or inf."\
                         f" You provided {Y}")
    # Normalise the data and move it to the origin
    mx = np.mean(X)
    my = np.mean(Y)

    sx = 0.5 * (np.max(X) - np.min(X))
    sy = 0.5 * (np.max(Y) - np.min(Y))

    x = (X - mx) / sx
    y = (Y - my) / sy

    # Construct design matrix and scatter matrix
    D = np.zeros((X.size, 6))
    D[:, 0] = x*x
    D[:, 1] = x*y
    D[:, 2] = y*y
    D[:, 3] = x
    D[:, 4] = y
    D[:, 5] = np.ones(X.size)

    S = D.T @ D

    # Construct contraint matrix
    C = np.zeros((6, 6))
    C[1, 1] = 1
    C[2, 0] = -2
    C[0, 2] = -2
    # Solve eigensystem
    # Break into blocks
    tmpA = S[:3, :3]
    tmpB = S[:3, 3:]
    tmpC = S[3:, 3:]
    tmpD = C[:3, :3]
    tmpE = np.linalg.inv(tmpC) @ tmpB.T

    eval_x, evec_x = np.linalg.eig(np.linalg.inv(tmpD) @ (tmpA - (tmpB@tmpE)))

    # Find the positive eigenvalue (as det(tmpD) < 0)
    idx = np.argmax(np.logical_and(np.real(eval_x) < 1e-8, np.isfinite(eval_x)))
    vec_x = np.real(evec_x[:, idx]) # vector associated with idx
    vec_y = -tmpE @ vec_x
    A = np.concatenate((vec_x, vec_y))

    # Un-normalise
    par = np.zeros(6)
    par[0] = A[0] * sy * sy
    par[1] = A[1] * sx * sy
    par[2] = A[2] * sx * sx
    par[3] = (-2*A[0]*sy*sy*mx) - (A[1]*sx*sy*my) + (A[3]*sx*sy*sy)
    par[4] = (-A[1]*sx*sy*mx) - (2*A[2]*sx*sx*my) + (A[4]*sx*sx*sy)
    par[5] = (A[0]*sy*sy*mx*mx) + (A[1]*sx*sy*mx*my) + (A[2]*sx*sx*my*my) \
                - (A[3]*sx*sy*sy*mx) - (A[4]*sx*sx*sy*my) + (A[5]*sx*sx*sy*sy)

    # Geometric radii and centres 
    theta_rad = 0.5*np.arctan2(par[1], par[0]-par[2])
    cos_t = np.cos(theta_rad)
    sin_t = np.sin(theta_rad)
    sin2 = sin_t * sin_t
    cos2 = cos_t * cos_t
    Ao = par[5]
    Au = (par[3] * cos_t) + (par[4] * sin_t)
    Av = (-par[3] * sin_t) + (par[4] * cos_t)
    Auu = (par[0] * cos2) + (par[2] * sin2) + (par[1] * cos_t * sin_t)
    Avv = (par[0] * sin2) + (par[2] * cos2) - (par[1] * cos_t * sin_t)

    tu_centre = -Au / (2*Auu)
    tv_centre = -Av / (2*Avv)
    w_centre = Ao - (Auu*tu_centre*tu_centre) - (Avv*tv_centre*tv_centre)

    x_centre = (tu_centre * cos_t) - (tv_centre * sin_t)
    y_centre = (tu_centre * sin_t) + (tv_centre * cos_t)

    Ru = -w_centre / Auu
    Rv = -w_centre / Avv

    Ru = np.sqrt(np.abs(Ru)) * np.sign(Ru)
    Rv = np.sqrt(np.abs(Rv)) * np.sign(Rv)

    return x_centre, y_centre, Ru, Rv, theta_rad



"""Tests for PlaceField"""
import numpy as np
import pytest

from opexebo.analysis import place_field as func
import opexebo.errors as err

print("=== tests_analysis_placeField ===")


###############################################################################
################                MAIN TESTS
###############################################################################

# Simple input validation
def test_invalid_input():
    fmap = np.ones((40, 40))
    with pytest.raises(err.ArgumentError):  # invalid input threshold
        init_thresh = 1.2
        func(fmap, init_thresh=init_thresh)
    with pytest.raises(err.ArgumentError):  # invalid input threshold
        init_thresh = 0
        func(fmap, init_thresh=init_thresh)
    with pytest.raises(err.ArgumentError):  # invalid search method - wrong type
        sm = 3
        func(fmap, search_method=sm)
    with pytest.raises(err.ArgumentError):  # invalid search method - unknown
        sm = "3"
        func(fmap, search_method=sm)
    with pytest.raises(
        NotImplementedError
    ):  # invalid search method - known but not implemented
        sm = "not implemented"
        func(fmap, search_method=sm)
    print("test_invalid_input() passed")


def test_unchanging_ratemap():
    """Check that the input arguments are not modified in place by the placefield detection function"""
    # All finite
    fmap = np.ones((40, 40))
    fmap_original = fmap.copy()
    func(fmap)
    assert np.array_equal(fmap, fmap_original)
    # Some non-finite
    fmap = np.ones((40, 40))
    fmap[21, 3] = np.inf
    fmap_original = fmap.copy()
    func(fmap)
    assert np.array_equal(fmap, fmap_original)
    # Masked array
    fmap = np.ones((40, 40))
    fmap[21, 3] = np.inf
    fmap = np.ma.masked_invalid(fmap)
    fmap_original = fmap.copy()
    func(fmap)
    assert np.array_equal(fmap, fmap_original)
    print("test_unchanging_ratemap() passed")



import numpy as np
import opexebo
import opexebo.defaults as default


array_like = (np.ndarray, np.ma.MaskedArray, list, tuple)

def speed_occupancy(time, speed, **kwargs):
    '''
    Calculate speed occupancy from speed
    
    Parameters
    ----------
    time : numpy.ndarray
        time stamps of angles in seconds
    speed : numpy.ndarray
        animal speed
    bin_width : float, optional
        Width of speed histogram bins
    
    Returns
    -------
    speed_histogram : numpy.ma.MaskedArray
        Speed histogram: masked array of time spent (in seconds) in specific
        speed ranges
    coverage: float
        Fraction of the speed ranges that the animal occupied. Range [0, 1]
    bin_edges: numpy.ndarray
        Bin edges of the speed histogram. len(bin_edges) == len(speed_histogram)+1
    '''
    for arg in (time, speed):
        if not isinstance(arg, array_like):
            raise ValueError("Time and Speed must be array-like")
        if not arg.ndim == 1:
            raise ValueError("Time and Speed must be 1d")
    if not time.size == speed.size:
        raise ValueError("Time and Speed must be the same size")
    
    bin_width = kwargs.get("bin_width", default.bin_speed)
    arena_size = max(speed)
    limits = (0, arena_size)
    
    speed_histogram, bin_edges = opexebo.general.accumulate_spatial(speed,
                                                                    bin_width=bin_width,
                                                                    arena_size=arena_size,
                                                                    limits=limits)
    masked_speed_histogram = np.ma.masked_where(speed_histogram==0, speed_histogram)
    
    frame_duration = np.mean(np.diff(time))
    masked_speed_histogram_seconds = masked_speed_histogram * frame_duration
    
    coverage = np.count_nonzero(speed_histogram) / masked_speed_histogram_seconds.size
    
    return masked_speed_histogram_seconds, coverage, bin_edges


def speed_map(time, speed, **kwargs):
    '''Calculate the firing rate of the animal as a function of its speed
    '''
    raise NotImplementedError
    
    
    


import os
import platform

"""Astropy insists on having a local temporary directory, and on on managed
NTNU computers, the default search location happens to be a non-responsive
network location with an excessively long (~1 minute) timeout. Therefore, fix
it to something harmless in advance"""
if "Windows" in platform.platform():
    os.environ["HOMESHARE"] = r"C:\temp\astropy"
elif "Linux" in platform.platform():
    os.environ["HOMESHARE"] = r"/tmp/astropy"
elif "macos" in platform.platform():
    os.environ["HOMESHARE"] = r"/tmp/astropy"


from .test_helpers import *
from . import test_analysis as analysis
from . import test_general as general



from .validate_keyword import validatekeyword__arena_size, validate_keyword_arena_shape
from .normxcorr2_general import normxcorr2_general
from .smooth import smooth
from .shuffle import shuffle
from .fit_ellipse import fit_ellipse
from .upsample import upsample
from .bin_width_2_num import bin_width_to_bin_number
from .circular_mask import circular_mask
from .accumulate_spatial import accumulate_spatial
from .peak_search import peak_search
from .power_spectrum import power_spectrum
from .spatial_cross_correlation import spatial_cross_correlation
from .walk_filter import walk_filter


__all__ = ['normxcorr2_general', 'smooth', 'accumulate_spatial', 'shuffle',
           'validatekeyword__arena_size', 'validate_keyword_arena_shape',
           'fit_ellipse', 'peak_search',
           'power_spectrum', 'spatial_cross_correlation', 'bin_width_to_bin_number',
           'circular_mask', 'upsample', 'walk_filter']



# Hack to speed up the borked Astropy configuration search
import os
import pathlib
os.environ["HOMESHARE"] = str(pathlib.Path.home())

import numpy as np
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # Suppress the ConfigurationMissingWarning that Astropy triggers
    import astropy.stats.circstats as cs

import opexebo.defaults as default
import opexebo.errors as error


def tuning_curve_stats(tuning_curve, **kwargs):
    """ Calculate statistics about a turning curve
    
    STATUS : EXPERIMENTAL

    Calculates various statistics for a turning curve.
    1. Mean vector length of a head direction rate map.
    The value will range from 0 to 1. 0 means that there are so much dispersion
    that a mean angle cannot be described. 1 means that all data are
    concentrated at the same direction. Note that 0 does not necessarily
    indicate a uniform distribution.
    Calculation is based on Section 26.4, J.H Zar - Biostatistical Analysis 5th edition,
    see eq. 26.13, 26.14.

    Parameters
    ----------
    tuning_curve : np.ma.MaskedArray
        Smoothed turning curve of firing rate as a function of angle
        Nx1 array
    kwargs
        percentile : float
            Percentile value for the head direction arc calculation
            Arc is between two points with values around
            globalPeak * percentile. Value should be in range [0, 1]

    Returns
    -------
    tcstat : dict
        hd_score         : float
            Score for how strongly modulated by angle the cell is
        hd_mvl           : float
            mean vector length
        hd_peak_rate     : float
            Peak firing rate  [Hz]
        hd_mean_rate     : float
            Mean firing rate [Hz]
        hd_peak_direction : float
            Direction of peak firing rate [degrees]
        hd_peak_direction_rad : float
            Direction of peak firing rate
        hd_mean_direction: float
            Direction of mean firing rate [degrees]
        hd_mean_direction_rad: float
            Direction of mean firing rate
        hd_stdev         : float
            Circular standard deviation [degrees]
        halfCwInd  : int
            Indicies of at the start, end of the range defined by percentile
            (clockwise).
        halfCcwInd : int
            Indicies of at the start, end of the range defined by percentile
            (counter-clockwise).
        halfCwRad : float
            Angle of the start, end of the range defined by percentile
        halfCcwRad  : float
            Angle of the start, end of the range defined by percentile
        arc_angle_rad : float
            Angle of the arc defined by percentile
        arc_angle_rad : float
            Angle of the arc defined by percentile

    Notes
    --------
    BNT.+analyses.tcStatistics

    Copyright (C) 2019 by Simon Ball

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    """

    debug = kwargs.get("debug", False)
    percentile = kwargs.get('percentile', default.hd_percentile)
    ndim = tuning_curve.ndim
    if ndim != 1:
        raise error.DimensionMismatchError(
            "tuning_curve should be a 1D array. You have provided {} dimensions".format(ndim)
            )
    if not 0 <= percentile <= 1:
        raise error.ArgumentError(
            "Keyword 'percentile' should be in the range [0, 1]. You provided  {:.2f.}".format(percentile)
            )
    if type(tuning_curve) != np.ma.MaskedArray:
        tuning_curve = np.ma.masked_invalid(tuning_curve)

    num_bin = tuning_curve.size
    bin_width = 2 * np.pi / num_bin
    hb = bin_width / 2
    bin_centres = np.linspace(hb, (2*np.pi)-hb, num_bin)

    if debug:
        print("Num_bin: %d" % num_bin)
        print("Therefore, bin_width = %.3g deg = %.3g rad"
              % (np.degrees(bin_width), bin_width))

    #### Calculate the simple values
    tcstat = {}
    # The average of the values of angles, weighted by the firing rate at those angles
    mean_dir_radians = cs.circmean(data = bin_centres, weights=tuning_curve)
    tcstat['hd_mean_direction_rad'] = mean_dir_radians
    tcstat['hd_mean_direction'] = np.degrees(mean_dir_radians)
    
    # The direction in which the highest firing rate occurs
    peak_dir_index = np.nanargmax(tuning_curve)
    peak_dir_angle_radians = _index_to_angle(peak_dir_index, bin_width)    
    tcstat['hd_peak_direction_rad'] = peak_dir_angle_radians
    tcstat['hd_peak_direction'] = np.degrees(peak_dir_angle_radians)
    
    # The peak firing rate IN Hz
    peak_rate_hz = np.nanmax(tuning_curve)
    tcstat['hd_peak_rate'] = peak_rate_hz
    
    # The mean firing rate across all angles IN Hz
    if tuning_curve.mask.all():
        #### Added to cope with numpy bug in nanmean with fully masked array
        mean_rate_hz = np.nan
    else:
        mean_rate_hz = np.nanmean(tuning_curve)
    tcstat['hd_mean_rate'] = mean_rate_hz


    #### Calculate the more complex ones:
    # mvl    
    mvl = np.sum(tuning_curve * np.exp(1j*bin_centres))
    mvl = np.abs(mvl)/np.sum(tuning_curve)
    tcstat['hd_mvl'] = mvl

    # hd_stdev
    # Eq. 26.20 from J. H. Zar
    tcstat['hd_stdev'] = np.sqrt(2*(1-mvl))

    # Percentile arc
    half_peak = peak_rate_hz * percentile

    # Because Python doesn't natively handle circular arrays, reshape such that
    # the peak rate occurs at the centre of the array - then don't have to worry
    # about whether the arc goes off one edge of the array or not
    # Must be careful to keep track of the array to which the indicies point
    tuning_curve_re = np.zeros_like(tuning_curve)
    centre_index = int(num_bin / 2)
    offset = centre_index - peak_dir_index
    tuning_curve_re = np.roll(tuning_curve, offset)
    # A positive offset means that the peak angle was in the range [0, pi], and
    # is now at the central index. Therefore, to get the "proper" index,
    # subtract offset from index in tuning_curve_re


    if debug:
        print("Centre index: %d, value" % centre_index)
        print("Peak index: %d" % peak_dir_index)
        print("Offset: %d" % offset)

    # Clockwise and counter-clockwise edges of arc around peak defined by
    # percentile. ccw index +1 to account for width of central peak
    cw_hp_index = np.where(tuning_curve_re >= (half_peak))[0][0] - offset
    ccw_hp_index = np.where(tuning_curve_re >= (half_peak))[0][-1] - offset+1

    cw_hp_ang = _index_to_angle(cw_hp_index, bin_width)
    ccw_hp_ang = _index_to_angle(ccw_hp_index, bin_width)
    arc_angle = ccw_hp_ang - cw_hp_ang

    if debug:
        print("CW: %d, %.3g rad" % (cw_hp_index, cw_hp_ang))
        print("CCW: %d, %.3g rad" % (ccw_hp_index, ccw_hp_ang))
        print("Arc: %.3g rad" % arc_angle)

    score = 1 - (arc_angle / np.pi)
    tcstat['halfCwInd'] = cw_hp_index
    tcstat['halfCcwInd'] = ccw_hp_index
    tcstat['halfCwRad'] = cw_hp_ang
    tcstat['halfCcwRad'] = ccw_hp_ang
    tcstat['arc_angle_rad'] = arc_angle
    tcstat['arc_angle_deg'] = np.degrees(arc_angle)
    tcstat['hd_score'] = score

    return tcstat


def _index_to_angle(index, bin_width):
    '''Given an index in the turning curve, return the angle at the centre of
    the bin that it indexes. The angle is deliberately not wrapped into [0,2pi]
    '''
    return (0.5 + index) * bin_width



""" Tests for spatial occupancy"""
import numpy as np
import matplotlib.pyplot as plt

from opexebo.analysis import border_score as func

print("=== tests_analysis_border_score ===")


###############################################################################
################                HELPER FUNCTIONS
###############################################################################


def field(grid, centre, width, peak):
    """
    grid: array
        the binned arena, e.g. np.zeros((40,40))
    centre: tuple
        x, y co-ordinates for the field to be centred
        (0,0) is assumed to be upper left
    width : tuple or float
        x, y sigma. If float, x=y
    peak : float
        peak rate
    """
    if type(width) in (float, int):
        width = (width, width)
    x = np.arange(grid.shape[1])
    y = np.arange(grid.shape[0])
    X, Y = np.meshgrid(x, y)
    xt = np.square(X - centre[0]) / (2 * width[0] ** 2)
    yt = np.square(Y - centre[1]) / (2 * width[1] ** 2)
    g = np.exp(-1 * (xt + yt))
    g[g < 0.025] = 0
    g *= peak
    return g


def show(f):
    plt.figure()
    plt.imshow(f)
    plt.show()


def rmap0():
    """f2 should have a border score of about 0.25 on the Right wall"""
    f = np.zeros((40, 40))
    f0 = field(f, (5, 14), (2.5, 4), 15)
    f1 = field(f, (23, 17), (1, 1.5), 10)
    f2 = field(f, (35, 3), (8, 8), 5)
    f3 = field(f, (12, 28), (6, 2), 6)
    f4 = field(f, (36, 37), (5, 3.3), 7)
    rmap0 = f0 + f1 + f2 + f3 + f4
    fmap0 = rmap0 > 3
    fields0 = [{"map": f0}, {"map": f1}, {"map": f2}, {"map": f3}, {"map": f4}]
    return rmap0, fmap0, fields0


def rmap1():
    """f0 should have a border score of 1 on the bottom wall"""
    f = np.zeros((40, 40))
    f0 = field(f, (20, 0), (40, 0.5), 15)
    f1 = field(f, (6, 32), (3, 3), 10)
    rmap1 = f0 + f1
    fmap1 = rmap1 > 3
    fields = [{"field_map": f0}, {"field_map": f1}]
    return rmap1, fmap1, fields


###############################################################################
################                MAIN TESTS
###############################################################################


def test_zero_fields():
    rmap = np.zeros((40, 40))
    fmap = rmap.copy()
    fields = []
    bs, cov = func(rmap, fmap, fields, arena_shape="s")
    assert bs == -1
    print("test_zero_fields() passed")


def test_perfect_fields():
    """Note: a perfect score (1) can only be achieved with a field that is full
    width and infinitesimal depth. Here we are testing with a 40x40 field, and
    a perfect score is 0.7288348...
    """
    bs, cov = func(*rmap1(), arena_shape="s", walls="B")
    assert bs >= 0.728 and bs <= 0.729


# if __name__ == '__main__':
#    test_zero_fields()
#    test_perfect_fields()
#    show(rmap1()[0])



""" Tests for spatial occupancy"""
import numpy as np
import pytest

from opexebo.analysis import border_coverage as func
from opexebo.analysis.border_coverage import _validate_keyword_walls as validate_func
from opexebo.errors import ArgumentError

print("=== tests_analysis_border_coverage ===")
kw_field_map = "field_map"

###############################################################################
################                validate_func
###############################################################################


def test_validate_func_square_correct_inputs():
    # Test correct strings
    # 1-4 characters, should yield the same string in lower case
    # Resulting string will be in alphabetical order and remove duplicates
    shape = "s"
    assert validate_func("TRBL", shape) == "blrt"
    assert validate_func("tBr", shape) == "brt"
    assert validate_func("b", shape) == "b"
    assert validate_func("bbbb", shape) == "b"
    assert validate_func("bbbTrBLl", shape) == "blrt"


def test_validate_func_square_incorrect_inputs():
    shape = "s"
    # zero length
    with pytest.raises(ValueError):
        validate_func("", shape)
    # >4 length of unique characters
    with pytest.raises(ValueError):
        validate_func("abcdef", shape)
    # With invalid characters
    with pytest.raises(ValueError):
        validate_func("abcd", shape)
    # angular array
    with pytest.raises(ArgumentError):
        validate_func([(15, 35), (270, 360)], shape)


###############################################################################
################                Square arena
###############################################################################


def test_perfect_fields_square():
    """Perfect field: 100% coverage of wall, zero distance, everywhere else zero"""
    shape = "s"
    perfect_left_field = np.zeros((40, 40))
    perfect_left_field[:, 0] = 1
    left = {kw_field_map: perfect_left_field}
    sw = 1
    assert func(left, shape, search_width=sw, walls="L") == 1

    right = {kw_field_map: np.fliplr(perfect_left_field)}
    assert func(right, shape, search_width=sw, walls="R") == 1

    top = {kw_field_map: np.rot90(perfect_left_field)}
    assert func(top, shape, search_width=sw, walls="T") == 1

    bottom = {kw_field_map: np.flipud(np.rot90(perfect_left_field))}
    assert func(bottom, shape, search_width=sw, walls="B") == 1

    # The above verify that the function agrees with a field that is there
    # The below verify that the function correctly identifies that a field is *not* there.
    assert func(left, shape, search_width=sw, walls="TRB") == 1 / 40
    print("test_perfect_fields() passed")


def test_offset_perfect_field_square():
    """Offset field: 100% coverage of wall, NaNs filling space between wall and field, distance 4, all else zero"""
    field = np.zeros((40, 40), dtype=float)
    field[:, 3] = 1
    field[:, :3] = np.nan
    field = np.ma.masked_invalid(field)
    fields = {kw_field_map: field}
    shape = "s"
    assert func(fields, shape, search_width=8, walls="L") == 1
    print("test_offset_field() passed")


def test_partial_field_square():
    """Fields that do not extend over the whole wall"""
    field = np.zeros((40, 40))
    field[:10, 0] = 1
    fields = {kw_field_map: field}
    walls = "L"
    assert func(fields, "s", walls=walls) == 0.25
    field[:20, 0] = 1
    assert func(fields, "s", walls=walls) == 0.5
    field[:30, 0] = 1
    assert func(fields, "s", walls=walls) == 0.75
    print("test_partial_field() passed")


def test_extended_field_square():
    """Field extending in two dimensions. Should get different answers depending
    on which wall direction is requested"""
    field = np.zeros((40, 40))
    field[:20, :30] = 1
    fields = {kw_field_map: field}
    assert func(fields, "s", walls="L") == 0.5
    assert func(fields, "s", walls="B") == 0.75
    assert func(fields, "s", walls="TRBL") == 0.75
    print("test_extended_field() passed")


def test_central_field_square():
    """A large firing field touching no wall"""
    field = np.zeros((40, 40))
    field[5:-5, 5:-5] = 1
    fields = {kw_field_map: field}
    assert func(fields, "s", search_width=8, walls="TRBL") == 0
    # The field is within the search width, BUT the search_wdth is about handling
    # locations that the animal never visited - the field MUST extend into the
    # location closest to the wall (that the animal visited) in order to qualify
    print("test_central_field() passed")


def test_zero_fields_square():
    fields = []
    assert func(fields, "s") == 0
    print("test_zero_fields() passed")



###############################################################################
################                Circular arenas
###############################################################################

# def test_perfect_field_circ():
#    plt.close("all")
#    num_bins = 410
#    axes = [np.linspace(-(num_bins-1), num_bins-1, num_bins) for i in range(2)]
#    diameter = 500#np.max(axes[0]) - np.min(axes[0])
#    in_field, _, _ = circular_mask(axes, diameter)
#    fields = {kw_field_map: in_field}
#    cov = func(fields, "circ", walls="tlbr", debug=True)
#
#    fig, ax = plt.subplots(1, 2)
#    ax[0].imshow(in_field)
#    print(cov)


#if __name__ == "__main__":
#    test_validate_func_square_correct_inputs()
#    test_validate_func_square_incorrect_inputs()
#    test_perfect_fields_square()
#    test_offset_perfect_field_square()
#    test_partial_field_square()
#    test_extended_field_square()
#    test_central_field_square()
#    test_zero_fields_square()
#    test_perfect_field_circ()



"""Provides a function for calculating a Rate map from an
Occupancy Map and positioned Spike Times"""

import numpy as np

import opexebo
#import opexebo.defaults as default
from .. import errors as err, defaults as default


def rate_map(occupancy_map, spikes_tracking, arena_size, **kwargs):
    ''' Calculate the spatially correlated firing rate map

    The rate map is calculated by the number of spikes in a bin divided by the
    time the animal spent in that bin. Bins in which the animal spent no or
    very little time are masked such that the value is available for but
    typically excluded from future analyses.

    The provided arena_size and bin_width **must* provide a number of bins such
    that the spike map has the same dimensions as the occupancy map.

    Parameters
    ----------
    occupancy_map: np.ndarray or np.ma.MaskedArray
        Nx1 or Nx2 array of time spent by animal in each bin, with time in bins
    spikes_tracking: np.ndarray
        Nx2 or Nx3 array of spikes tracking: [t, x] or [t, x, y].
    arena_size: float or tuple of floats
        Dimensions of arena (in cm)
            * For a linear track, length
            * For a circular arena, diameter
            * For a rectangular arena, length or (length, length)
    bin_width: float
        Bin size in cm. Default 2.5cm. If bin_width is supplied, `limit` must
        also be supplied. One of `bin_width`, `bin_number`, `bin_edges` must be
        provided
    bin_number: int or tuple of int
        Number of bins. If provided as a tuple, then `(x_bins, y_bins)`. One
        of `bin_width`, `bin_number`, `bin_edges` must be provided
    bin_edges: array-like
        Edges of the bins. Provided either as `edges` or `(x_edges, y_edges)`.
        One of `bin_width`, `bin_number`, `bin_edges` must be provided
    limits: tuple or np.ndarray
        (x_min, x_max) or (x_min, x_max, y_min, y_max)
        Provide concrete limits to the range over which the histogram searches
        Any observations outside these limits are discarded
        If no limits are provided, then use np.nanmin(data), np.nanmax(data)
        to generate default limits.
        As is standard in python, acceptable values include the lower bound
        and exclude the upper bound

    Returns
    -------
    rmap: np.ma.MaskedArray
        2D array, masked at locations of very low occupancy (t<1ms).
        Each cell gives the rate of neuron firing at that location.

    See Also
    --------
    opexebo.general.accumulate_spatial
    opexebo.general.bin_width_to_bin_number

    Notes
    --------
    BNT.+analyses.map()

    Copyright (C) 2019 by Simon Ball
    '''

    # Check correct inputs
    if not isinstance(occupancy_map, (np.ndarray, np.ma.MaskedArray)):
        raise err.ArgumentError("Occupancy Map not provided in usable format. Please"\
            " provide either a Numpy ndarray or Numpy MaskedArray. You"\
            f" provided {type(occupancy_map)}.")
    if not isinstance(spikes_tracking, (np.ndarray, np.ma.MaskedArray)):
        raise err.ArgumentError("spikes not provided in usable format. Please"\
            " provide either a Numpy ndarray or Numpy MaskedArray. You"\
            f" provided {type(spikes_tracking)}.")

    dims_p = occupancy_map.ndim
    dims_s, _ = spikes_tracking.shape
    if dims_s-1 != dims_p:
        raise err.DimensionMismatchError("Spikes must have the same number of spatial"\
            " dimensions as positions ([t, x] or [t, x, y]). You have provided"\
            f" {dims_s} columns of spikes, and {dims_p} columns of positions")

    if isinstance(occupancy_map, np.ndarray):
        occupancy_map = np.ma.MaskedArray(occupancy_map)


    if dims_s == 2:
        spikes = spikes_tracking[1, :]
    elif dims_s == 3:
        spikes_x = spikes_tracking[1, :]
        spikes_y = spikes_tracking[2, :]
        spikes = np.array((spikes_x, spikes_y))

    # Histogram of spike positions
    spike_map, _ = opexebo.general.accumulate_spatial(spikes, arena_size, **kwargs)

    if spike_map.shape != occupancy_map.shape:
        raise err.DimensionMismatchError("Rate Map and Occupancy Map must have the same"\
                         f" dimensions. Provided: {spike_map.shape},"\
                         f" {occupancy_map.shape}")
    # Convert to rate map
    rmap = spike_map / (occupancy_map + np.spacing(1))
    # spacing adds floating point precision to avoid DivideByZero errors
    # These should be impossible due to masking, but included nevertheless
    return rmap



""" Tests for keyword validation"""
import numpy as np
import pytest

from opexebo.general import validatekeyword__arena_size as func
from opexebo import errors as err

print("=== tests_general_validatekeyword ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_invalid_inputs():
    # >2D dimensional data
    with pytest.raises(err.ArgumentError):  # Negative length, 1d
        kwv = -1
        dim = 1
        func(kwv, dim)
    with pytest.raises(err.ArgumentError):  # Negative length, 2d
        kwv = -1
        dim = 2
        func(kwv, dim)
    with pytest.raises(err.ArgumentError):  # zero length
        kwv = 0
        dim = 1
        func(kwv, dim)
    with pytest.raises(NotImplementedError):  # Invalid # dimensions
        kwv = 80
        dim = 3
        func(kwv, dim)
    with pytest.raises(NotImplementedError):  # Invalid # dimensions
        kwv = 80
        dim = 0
        func(kwv, dim)
    with pytest.raises(err.DimensionMismatchError):  # Invalid # dimensions
        kwv = (80, 80)
        dim = 1
        func(kwv, dim)
    print("test_invalid_inputs passed")


def test_1d_int_valid():
    kwv = 80
    dim = 1
    ars, is_2d = func(kwv, dim)
    assert ars == 80.0
    assert not is_2d
    print("test_1d_int_valid passed")


def test_1d_str_valid():
    kwv = "80"
    dim = 1
    ars, is_2d = func(kwv, dim)
    assert ars == 80.0
    assert not is_2d
    print("test_1d_str_valid passed")


def test_2d_square_valid():
    kwv = 80
    dim = 2
    ars, is_2d = func(kwv, dim)
    assert isinstance(ars, np.ndarray)
    assert ars[0] == ars[1] == kwv
    assert is_2d
    print("test_2d_square_valid passed")


def test_2d_rectangle_valid():
    kwv = [80, 120]
    dim = 2
    ars, is_2d = func(kwv, dim)
    assert np.array_equal(ars, np.array((80, 120)))
    assert is_2d
    print("test_2d_rectangle_valid passed")



import numpy as np

import opexebo.errors as errors


def shuffle(
    times: np.ndarray,
    offset_lim: float,
    iterations: int,
    t_start: float = None,
    t_stop: float = None,
):
    """
    Duplicate the provided time series ``iterations`` number of times. Each
    duplicate will be incremented circularly by a random value not smaller than
    ``offset_lim``.

    Circular incrementation results in (the majority) of time _differences_
    remaining preserved

    * Initially, we have a time series, ``times``, both with
      values in the range ``[min(times), max(times)]``. ``t_start`` may be smaller than
      ``min(times)``, and ``t_stop`` may be larger than ``max(times)``
    * ``iterations`` number of duplicates of ``times`` are created.
    * In each iteraction, a random increment ``T`` is generated, and added to
      each value in that iteration, such that values now fall into the range
      ``[min(times)+T, max(times)+T]``. ``max(times)+T`` may exceed ``t_stop``.
    * All timestamps matching ``t_n > t_stop`` are mapped back into the range
    ``[t_start, t_stop]`` by subtracting ``(t_stop-t_start)``
    * The iteration is re-ordered by value (moving those beyond the far edge
      back to the beginning)

    Parameters
    ----------
    times : np.ndarray
        1D array of floats. Time series data to be shuffled
    offset_lim : float
        Minimum offset from the original time values. Each iteration is
        incremented by a random value evenly distributed in the range
        ``[offset_lim, t_stop-offset_lim]``
    iterations : int
        Number of repeats of ``times`` to be returned
    t_start : float, optional
        Lower bound of time domain. Must meet the criteria ``t_start <= min(times)``
        Defaults to ``min(times)``
    t_stop : float, optional
        Upper bound of time domain. Must meet the criteria ``t_stop >= max(times)``
        Defaults to ``max(times)``

    Returns
    -------
    output : np.ndarray
        iterations x N array of times. A single iteration is accessed as
        ``output[i]``
    increments : np.ndarray
        1D array of offset values that were used
    """
    # Argument checking begins here
    if not isinstance(times, np.ndarray):
        raise errors.ArgumentError(
            "`times` must be 1 Numpy array ({})".format(type(times))
        )
    if not times.ndim == 1:
        raise errors.ArgumentError("`times` must be a 1D array ({})".format(times.ndim))
    if not np.isfinite(times).all():
        raise errors.ArgumentError("`times` cannot include non-finite or NaN values")

    if offset_lim <= 0:
        raise errors.ArgumentError(
            "`offset_lim` must be greater than zero ({}".format(offset_lim)
        )
    if not np.isfinite(offset_lim):
        raise errors.ArgumentError(
            "`offset_lim` must be finite ({})".format(offset_lim)
        )

    if iterations < 2:
        raise errors.ArgumentError(
            "qiterations must be a positive integer greater than 1 ({})".format(
                iterations
            )
        )
    if not np.isfinite(iterations):
        raise errors.ArgumentError("`iterations` must be finite".format(iterations))

    if not np.isfinite(t_start):
        raise errors.ArgumentError("`t_start` must be finite".format(t_start))
    if not np.isfinite(t_stop):
        raise errors.ArgumentError("`t_stop` must be finite".format(t_stop))

    if t_start is None:
        t_start = min(times)
    if t_stop is None:
        t_stop = max(times)

    if t_start > min(times):
        raise errors.ArgumentError(
            "`t_start` must be greater than or equal to `min(times)`"
        )
    if t_stop < max(times):
        raise errors.ArgumentError(
            "`t_stop` must be less than or equal to `max(times)`"
        )

    if t_start == t_stop:
        raise errors.ArgumentError(
            "`t_start` and `t_stop` cannot be identical ({})".format(t_start)
        )

    if offset_lim >= 0.5 * (t_stop - t_start):
        raise errors.ArgumentError(
            "`offset_lim` must be less than half of the time span ({}, {})".format(
                offset_lim, t_stop - t_start
            )
        )
    # argument checking ends here

    # Main logic begins here
    increments_base = np.random.RandomState().rand(
        iterations
    )  # uniformly distributed in [0,1]
    increments = (
        t_start + offset_lim + (increments_base * (t_stop - t_start - 2 * offset_lim))
    )

    # Stack copies of `times`, one per row, for `iterations` number of rows
    # stack copies of `increments`, one per column, for `times.size` number of columns
    # We get two identically shaped arrays that can just be added together to perform the increments.
    output = np.repeat(times[np.newaxis, :], iterations, axis=0)
    increments_arr = np.repeat(increments[:, np.newaxis], times.size, axis=1)

    output = increments_arr + output

    # Circularising: i.e. folding times outside the boundary back inside the
    # boundary, and then re-ordering by the updated, refolded times
    out_of_bounds = output > t_stop
    output[out_of_bounds] = output[out_of_bounds] - (t_stop - t_start)

    output.sort(axis=1)  # sort along each row independently of all other rows.

    return output, increments



"""Provide function for calculating a Border Score"""

import numpy as np
from scipy.ndimage import distance_transform_edt

import opexebo.defaults as default
from opexebo.general import validate_keyword_arena_shape, circular_mask
from opexebo.errors import ArgumentError


def border_coverage(fields, arena_shape, **kwargs):
    '''
    Calculate border coverage for detected fields.
    
    STATUS : EXPERIMENTAL
    
    This function calculates firing map border coverage that is further
    used in calculation of a border score. Coverage is calculated on a per-field
    basis, and is therefore dependent on the exact details of the place-field
    detection algorithm. 
    
    Additionally, please be aware of what `top` and `bottom` mean. This function
    treats `top` as the location where the `y` indicies are greatest. This is
    the standard convention in a GRAPH ((0,0) at bottom left, `x` increases to
    right and `y` increases to top). Images are conventionally plotted UPSIDE-
    DOWN, with `y` increasing towards the bottom.

    Parameters
    ----------
    fields: dict or list of dicts
        One dictionary per field. Each dictionary must contain the keyword field_map
    arena_shape: {"square", "rect", "circle", "line"}
        Rectangular and square are equivalent. Elliptical or n!=4 polygons
        not currently supported.
    
    Other Parameters
    ----------------
    search_width: int
        rate_map and fields_map have masked values, which may occur within the
        region of border pixels. To mitigate this, we check rows/columns within
        search_width pixels of the border Default `8`
    walls: str or list of tuple
        Definition of which walls to consider for border coverage. Behaviour is
        different for different arena shapes
        * Rectangular arenas
          type str. The four walls are referred to as `T`, `B`, `R`, `L` for the
          standard cardinals. Case insensitive. You may include or exclude any
        * Circular arenas
          Walls should be specified as a list of inclusive angular start-stop pairs. Angles
          given in degrees clockwise from top. Both `0` and `360` are permissible
          values. Start-stop may wrap through zero and may be overlapping with
          other pairs. Negative angles may be used to express counter-clockwise
          rotation, and will be automatically wrapped into the [0, 360] range.
          Example: [(0, 15), (180, 270), (315, 45), (30, 60)]
          Walls may ALTERNATIVELY be expressed as a string, equivalent to rectangular
          arenas. In this case, walls will be interpreted as follows
              * `T` - (315, 45)
              * `B` - (135, 225)
              * `L` - (45, 135)
              * `R` - (225, 315)
        In either case, defaults `TRBL`. Converting this (or other string) for
        circluar arrays is handled internally

    Returns
    -------
    coverage: float
        Border coverage, ranges from 0 to 1.

    See Also
    --------
    opexebo.analysis.placefield
    opexebo.analysis.borderscore

    Notes
    --------
    * BNT.+analyses.placefield
    * BNT.+analyses.borderScore
    * BNT.+analyses.borderCoverage

    Copyright (C) 2019 by Simon Ball
    '''
    # Process input arguments
    debug = kwargs.get("debug", False)
    arena_shape = validate_keyword_arena_shape(arena_shape)
    kw_field_map = "field_map"
    
    if arena_shape in default.shapes_square:
        calculate_coverage = _calculate_coverage_rect
    elif arena_shape in default.shapes_circle:
        calculate_coverage = _calculate_coverage_circ
    else:
        raise NotImplementedError(f"Border coverage is not implemented for arena shape `{arena_shape}`")
    
    if isinstance(fields, dict):
        # Deal with the case of being passed a single field, instead of a list of fields
        fields = [fields]
    elif not isinstance(fields, (list, tuple, np.ndarray)):
        raise ValueError(f"You must supply either a dictionary, or list of dictionaries, of fields. You provided type '{type(fields)}'")
    for i, field in enumerate(fields):
        if kw_field_map not in field.keys():
            raise KeyError(f"field dictionary {i} does not have keyword '{kw_field_map}'.")

    # Extract keyword arguments or set defaults
    search_width = kwargs.get('search_width', default.search_width)
    walls = kwargs.get('walls', default.walls)

    # Perform validation of the `walls` argument
    walls = _validate_keyword_walls(walls, arena_shape)

    if debug:
        print("===== border_coverage =====")
        print(f"arena_shape: {arena_shape}")
        print(f"coverage method: {calculate_coverage}")
        print(f"walls: {walls}")
        print(f"num fields: {len(fields)}")

    coverage = 0
    for field in fields:
        fmap = field[kw_field_map]
        for wall in walls:
            c = calculate_coverage(fmap, wall, search_width, debug)
            coverage = max(coverage, c)
    
    return coverage


###############################################################################
####            Helper functions : CIRCULAR arenas
###############################################################################



def _calculate_coverage_circ(fmap, wall, search_width, debug):
    '''
    Evaluate what fraction of the border area is covered by a single field in a
    circular arena
    
    Problems to solve:
        * Identify where the border actually is, since we can't just use the
          edge of the array
        * Convert the cartesian co-ordinates of the field map to polar co-ordinates
        * Isolate the relevant arc
    First point: we can _start_ by assuming that the circle is centred on the
    centre of the array. I'm not convinced that this is a good approximation, but
    it's a place to start
    
    Parameters
    ----------
    fmap : np.ndarray or np.ma.MaskedArray
        Binary map of specific firing field
    '''
    if not isinstance(fmap, np.ma.MaskedArray):
        fmap = np.ma.asanyarray(fmap)
    fmap = fmap.copy() # create a separate copy to work on, to avoid propagating changes back outside this function
    
    # Create a border wall mask, and a distance map
    # use cv2 to 
    raise NotImplementedError("Border Coverage is not currently functional for circular arenas")
    # Below - using cv2 to find the minimum enclosing circular contour
    # The trouble is this only works on the _entire_ ratemap (and it seems to work fairly well)
    # It won't work at all on single fields. 
    # At least it would need to be run on the entire, summed, fields_map. 
    # Better still, it would be run on the ratemap. 
    
    fmap_threshold = np.uint8(fmap>0)
    centres = cv2.findContours(fmap_threshold, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2]
    x = y = radius = 0
    for c in centres:
        (x_, y_), r_ = cv2.minEnclosingCircle(c)
        if r_ > radius:
            x = x_
            y = y_
            radius = r_
    axes = [np.arange(0, fmap.shape[i]+1, 1) for i in [0,1]]
    circle_mask, distance, angle = circular_mask(axes, radius*2, origin=(x, y))
    
    if debug:
        print(f"radius (bins): {radius}")
    
    
    

    
    # Circular_mask gives us nearly everything we need: an array of distances (in floating-point bins)
    # and a separate array of angles (in degrees, with 0 pointing to y_max, and 90° pointing to x_max)
    # We can use these as lookups to convert to polar co-ordinates
    
    
    # Construct a mask for locations we do care about, based on angle, 
    # distance_from_border, and whether it is inside the circle
    if wall[0] > wall[1]:
        good_arc = np.logical_or(np.logical_and(wall[0] <= angle,
                                                angle <= 360),
                                 np.logical_and(0 <= angle,
                                                angle <= wall[1]))
    else:
        good_arc = np.logical_and(wall[0] <= angle,
                                  angle <= wall[1])
    
    border_distance = radius - distance
    good_distance = np.logical_and(0 <= border_distance,
                                   border_distance <= search_width)
    
    in_region = np.logical_and(circle_mask, good_arc, good_distance)
    not_in_region = np.logical_not(in_region)
    
    # Using this mask, blank out all the locations that are not relevant
    # "Blanking out" the border_distance array , means set it to high values, not low
    fmap[not_in_region] = 0
    border_distance[not_in_region] = np.max(border_distance)
    
    # Plan
        # Iterate across all locations within `in_region`, low to high, sorted by `border_distance`
        # Locations which round (floor) down to zero will potentially count towards coverage
            # To count towards coverage, they must be part of a field
            # Alternatively, if they are NaN in `fmap` (never visited), then find the next nearest in along that angle, and repeat the check




    def _circular_inwards_search(idx, i):
        '''
        If an unvisited location is found, recursively search inwards until either
        search_width is reached, or a visited location is found. If a visited location
        is found, at nearly the same angle, then swap the values in field_map
        
        This recursive loop is only started where `border_distance_int`==0, so start
        searching where `border_distance_int` == 1, and recursively allow it to rise, 
        until the value at the closest angle is NOT a NaN
        '''
        nonlocal fmap
        nonlocal border_distance
        nonlocal border_distance_int
        nonlocal angle
        nonlocal in_region
        nonlocal search_width
        
        if i == search_width:
            # Recurse at most `search_width` times
            return None
        
        locations_to_search = (border_distance_int == i)
        try:
            idx_2 = np.argmin(np.abs(angle[locations_to_search] - angle[idx]))
        except ValueError:
            # if the sequence is empty, try the next i
            _circular_inwards_search(idx, i+1)
        if np.isnan(fmap[idx_2]):
            # If this location is NaN as well, try the next i
            _circular_inwards_search(idx, i+1)
        else:
            # If not NaN, then swap the values, and return to the main loop
            fmap[idx] = fmap[idx_2]


    
    border_distance_int = np.floor(border_distance)
    # sort by border_distance and then work with co-ordinates to index between the three arrays (fmap, border_distance, angle)
    locations_to_search =  np.argwhere(border_distance_int == 0)
    if debug:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(ncols=2)
        ax[0].imshow(fmap)
        ax[1].imshow(border_distance_int)

    for row, col in locations_to_search:
        idx = (row, col)
        if np.isnan(fmap[idx]):
            # Do some shuffling
            # TODO
            _circular_inwards_search(idx, 1)
            pass
        if fmap[idx]:   # i.e. is part of the field, after shuffling
            # do nothing, leave it as a zero for the final count
            pass
        else:           # i.e. not part of the field
            # Set the value in border_distance_int to non-zero, so it will not be counted
            border_distance_int[idx] = np.max(border_distance_int)
    
    # Coverage is determined by the ratio of locations with distance 0 to the number covered by the field in question    
    coverage = np.count_nonzero(border_distance_int==0) / np.max(locations_to_search.shape)
            
    return coverage
    
    
    
###############################################################################
####            Helper functions : RECTANGULAR arenas
###############################################################################


def _calculate_coverage_rect(fmap, wall, search_width, debug):
    '''
    Evaluate what fraction of the border area is covered by a single field in a
    rectangular or square arena

    Border coverage is provided as two values: 
        covered : the sum of the values across all sites immediately adjacent 
            to the border, where the values are calculated from the distance of
            those sites to the firing field, excluding NaN, inf, and masked values
        norm    : the number of non nan, inf, masked values considered in the 
            above sum

    The border area is defined by wfmap - this is the subsection of the binary 
    firing map of a  single field that lies within search_width of a border. 
    wfmap is must be of size NxM where
        N : arena_size / bin_size
        M : search_width
    and where the 0th column (wfmap[:,0], len()=N) repsents the sites closest 
    to the border

    fmap: has value 1 inside the field and 0 outside the field
    
    Parameters
    ----------
    fmap : np.ndarray or np.ma.MaskedArray
        Binary map of a sngle firing field (as supplied by `opexebo.analysis.place_field`).
    wall : str
        Which wall to consider for coverage. Call this function multiple times
        to consider multiple walls
    search_width : int
        Width of the region from the border inwards to consider for calculation
        
    
    Returns
    -------
    coverage : float
        Coverage of the relevant border. Coverage is maximised for an infinitely
        narrow field extending over the entire width of the border. Thicker fields
        that cover less of the border width have reduced coverage. 
    '''
    # fmap is the input field
    # wfmap is the isolated and rotated section to search
    wfmap = _get_aux_map_rect(fmap, wall, search_width)
    
    if type(wfmap) != np.ma.MaskedArray:
        wfmap = np.ma.asanyarray(wfmap)
    wfmap[np.ma.greater(wfmap, 1)] = 1 # Just in case a still-labelled map has crept in
    inverted_wfmap = 1 - wfmap 
    distance = distance_transform_edt(np.nan_to_num(inverted_wfmap.data, copy=True)) # scipy doesn't recognise masks
    distance = np.ma.masked_where(np.ma.getmaskarray(wfmap), distance) # Preserve masking
    # distance_transform_edt(1-map) is the Python equivalent to (Matlab bwdist(map))
    # Cells in map with value 1 go to value 0
    # Cells in map with value 0 go to the geometric distance to the nearest value 1 in map

    adjacent_sites = distance[:,0]
    # Identify sites which are NaN, inf, or masked
    # Replace them with the next cell along the row, closest to the wall, that is not masked, nan, or inf
    adjacent_sites.mask += np.isnan(adjacent_sites.data)
    if adjacent_sites.mask.any():
        for i, rep in enumerate(adjacent_sites.mask):
            if rep:
                for j, val in enumerate(distance[i,:]):
                    if not distance.mask[i,j] and not np.isnan(val):
                        adjacent_sites[i] = val
                        adjacent_sites.mask[i] = False
                        break
    covered = np.ma.sum(adjacent_sites==0)
    contributing_cells = wfmap.shape[0] - np.sum(adjacent_sites.mask) # The sum gives the number of remaining inf, nan or masked cells
    coverage = covered / contributing_cells
    return coverage



def _get_aux_map_rect(fmap, wall, search_width):
    '''
    Isolate the appropriate section of a map (determined by `search_width`), and
    rotate it so that the relevant portion, and the border are at left most 
    (i.e. minimum `x` index)
    
    !! WARNING !! This function treats "top" as "the values where the `y` index is greatest.
    THIS DOES NOT NECESSARILY MEAN THE TOP OF THE IMAGE!
    
    Conventionally, a graph is plotted with (x=0, y=0) at BOTTOM left. X increases
    towards the right, and Y increases towards the top
    
    Conventionally an image is plotted flipped up/down, with (0,0) at TOP left,
    and Y increases towards the BOTTOM
    '''
    # Get our "auxiliary" map from the provided firing map
    # That means pick out the particular section determined by `search_width`
    # And rotate as determined by `wall`, so that the relevant border and pixels 
    # are always at the left-most of the image
    
    if wall == "l":
        aux_map = fmap[:,:search_width].copy()
    elif wall == "r":
        # Mirror image to match the expectations in _wall_field, i.e. border adjacent to left-most column
        aux_map = fmap[:, -search_width:].copy()
        aux_map = np.fliplr(aux_map)
    elif wall == "t":
        # rotate 90 deg counter clockwise (bottom to right), then mirror image
        aux_map = fmap[-search_width:, :].copy()
        aux_map = np.rot90(aux_map)
        aux_map = np.fliplr(aux_map)
    elif wall == "b":
        # Rotate counterclockwise - top of image moves to left of image
        aux_map = fmap[:search_width, :].copy()
        aux_map = np.rot90(aux_map)
    else:
        # This should never happen
        raise ValueError
    return aux_map





###############################################################################
####            Helper functions : Misc
###############################################################################

def _validate_keyword_walls(walls, shape):
    '''Parse the walls keyword to be sure that it is valid and conforms to requirements.
    Logic is complicated due to the wildly diverging approach of the square and circular arena shapes
    
    Parameters
    ----------
    walls : keyword value given to border_coverage `walls`
    shape : keyword value given to border_coverage `arena_shape`
    
    Returns
    -------
    walls : str or list of np.ndarrays
        Appropriate form given the `shape` parameter
    
    '''
    
    if isinstance(walls, str):
        # do string-related stuff
        # Check that only acceptable characters are included
        # Remove duplicates and set to lowercase
        walls = "".join(sorted(set(walls.lower())))
        # Note: set() can break ordering. sorting still breaks ordering, but makes it consistent
        if not (0 < len(walls) <= 4):
            raise ValueError("keyword `walls` as a string must contain 1-4 characters."\
                             f" You provided {len(walls)} characters")
        for char in walls:
            if char not in ["t","r","b","l"]:
                 raise ValueError("Character %s is not a valid entry in wall"\
                                  " definition. Valid characters are [t, r, b, l]" % char)
        # If the arena_shape is circular, convert this to the requisite angular form
        if shape in default.shapes_circle:
            walls_new = []
            for char in walls:
                if char == "t":
                    walls_new.append((315, 45))
                elif char == "b":
                    walls_new.append((135, 225))
                elif char == "l":
                    walls_new.append((45, 135))
                elif char == "r":
                    walls_new.append((225, 315))
            walls = walls_new

    elif isinstance(walls, (list, tuple, np.ndarray)):
        if shape in default.shapes_square:
            raise ArgumentError("Keyword `walls` is the wrong type for a rectangular"\
                                " arena. You must provide a string definition")
        for i, element in enumerate(walls):
            if not isinstance(element, (list, tuple, np.adarray)):
                raise ValueError("Keyword `walls` as an array-like must contain array-like"\
                                 f" start-stop pairs. Element {i} has type `{type(element)}`")
            if not len(element) == 2:
                raise ValueError("Keyword `walls` as an array-like must contain array-like"\
                                 f" start-stop pairs. Element {i} has length `{len(element)}`")
            # Ensure that values are wrapped into [0, 360]
            # Ensure that values are ordered (smaller, larger)
            new_element = np.array(sorted(element)) % 360
            walls[i] = new_element

    else:
        raise ArgumentError("Keyword `wall` definition is the wrong type. You"\
                            "must provide a string (for rectangular or circular arrays),"\
                            " or a list of angular start-stop lists for circular arrays")

    return walls


"""
Python version of Matlab's normxcorr2_general

This is a Python adaption of code found at
https://se.mathworks.com/matlabcentral/fileexchange/29005-generalized-normalized-cross-correlation
Since we use it for autocorrelograms exclusively some input arguments of the
original function have been dropped.
"""

from __future__ import division

import numpy as np
from scipy.signal import convolve2d

REQUIRED_OVERLAP_PIXELS = 0


def normxcorr2_general(array):
    """Calculate spatial autocorrelation.

    Python implementation of the Matlab `generalized-normalized cross correlation`
    function, adapted by Vadim Frolov. Some generality was abandoned in the adaption
    as unnecessary for autocorrelogram calculation.
    
    For the original function, see https://se.mathworks.com/matlabcentral/fileexchange/29005-generalized-normalized-cross-correlation

    Parameters
    ----------
    array: NxM matrix
        firing array. array is not necessary a numpy array. Must not contain NaNs!

    Returns
    -------
    np.ndarray
        Resulting correlation matrix
    """

    if not isinstance(array, np.ndarray):
        array = np.array(array)
    if not np.sum(np.isfinite(array)) == array.size:
        raise ValueError("Input array contains NaN values.")

    
    A = _shift_data(array)
    T = _shift_data(array)

    number_of_overlap_pixels = _local_sum(np.ones(A.shape), T.shape[0], T.shape[1])

    local_sum_A = _local_sum(A, T.shape[0], T.shape[1])
    local_sum_A2 = _local_sum(A*A, T.shape[0], T.shape[1])

    # Note: diff_local_sums should be nonnegative, but it may have negative
    # values due to round off errors. Below, we use max to ensure the radicand
    # is nonnegative.
    diff_local_sums_A = (local_sum_A2 - np.power(local_sum_A, 2) / number_of_overlap_pixels)
    del local_sum_A2

    denom_A = np.maximum(diff_local_sums_A, 0)
    del diff_local_sums_A

    # Flip T in both dimensions so that its correlation can be more easily
    # handled.
    rotatedT = np.rot90(T, 2)
    local_sum_T = _local_sum(rotatedT, A.shape[0], A.shape[1])
    local_sum_T2 = _local_sum(rotatedT*rotatedT, A.shape[0], A.shape[1])
    del rotatedT

    diff_local_sums_T = (local_sum_T2 - np.power(local_sum_T, 2) / number_of_overlap_pixels)
    del local_sum_T2
    denom_T = np.maximum(diff_local_sums_T, 0)
    del diff_local_sums_T

    denom = np.sqrt(denom_T * denom_A)
    del denom_T, denom_A

    xcorr_TA = _xcorr2_fast(T, A)
    del A, T
    numerator = xcorr_TA - local_sum_A * local_sum_T / number_of_overlap_pixels
    del xcorr_TA, local_sum_A, local_sum_T

    # denom is the sqrt of the product of positive numbers so it must be
    # positive or zero.  Therefore, the only danger in dividing the numerator
    # by the denominator is when dividing by zero. We know denom_T~=0 from
    # input parsing; so denom is only zero where denom_A is zero, and in these
    # locations, C is also zero.
    C = np.zeros(numerator.shape)
    tol = 1000 * np.spacing(np.max(np.abs(denom)))
    i_nonzero = (denom > tol)
    C[i_nonzero] = numerator[i_nonzero] / denom[i_nonzero]
    del numerator, denom

    # Remove the border values since they result from calculations using very
    # few pixels and are thus statistically unstable.
    # By default, REQUIRED_OVERLAP_PIXELS = 0, so C is not modified.
    if REQUIRED_OVERLAP_PIXELS > np.max(number_of_overlap_pixels):
        raise ValueError("ERROR: REQUIRED_OVERLAP_PIXELS")

    C[number_of_overlap_pixels < REQUIRED_OVERLAP_PIXELS] = 0
    return C


def _xcorr2_fast(T, A):
    T_size = T.shape
    A_size = A.shape
    outsize = np.array(A.shape) + np.array(T.shape) - 1

    # Figure out when to use spatial domain vs. freq domain
    conv_time = _time_conv2(T_size, A_size)  # 1 conv2
    fft_time = 3*_time_fft2(outsize)  # 2 fft2 + 1 ifft2

    if conv_time < fft_time:
        cross_corr = convolve2d(np.rot90(T, 2), A)
    else:
        cross_corr = _freqxcorr(T, A, outsize)

    return cross_corr


def _freqxcorr(a, b, outsize):
    # Find the next largest size that is a multiple of a combination of 2, 3,
    # and/or 5.  This makes the FFT calculation much faster.
    optimalSize = np.zeros((2, 1))
    optimalSize[0] = _find_closest_valid_dimension(outsize[0])
    optimalSize[1] = _find_closest_valid_dimension(outsize[1])
    optimalSize = optimalSize.squeeze()
    optimalSize = optimalSize.astype(int)

    # Calculate correlation in frequency domain
    rot_version = np.rot90(a, 2)
    Fa = np.fft.fft2(rot_version, s=(optimalSize[0], optimalSize[1]))
    Fb = np.fft.fft2(b, s=(optimalSize[0], optimalSize[1]))
    xcorr_ab = np.real(np.fft.ifft2(Fa * Fb))

    xcorr_ab = xcorr_ab[0:outsize[0], 0:outsize[1]]
    return xcorr_ab


def _time_conv2(obssize, refsize):
    # K was empirically calculated by the commented-out code above.
    K = 2.7e-8

    # convolution time = K*prod(obssize)*prod(refsize)
    time = K * np.prod(obssize) * np.prod(refsize)
    return time


def _time_fft2(outsize):
    # time a frequency domain convolution by timing two one-dimensional ffts

    R = outsize[0]
    S = outsize[1]

    # Tr = time_fft(R)
    # K_fft = Tr/(R*log(R))

    # K_fft was empirically calculated by the 2 commented-out lines above.
    K_fft = 3.3e-7
    Tr = K_fft * R * np.log(R)

    if S == R:
        Ts = Tr
    else:
        # Ts = time_fft(S)  % uncomment to estimate explicitly
        Ts = K_fft * S * np.log(S)

    time = S*Tr + R*Ts
    return time


def _local_sum(A, m, n):
    """
    This algorithm depends on precomputing running sums.

    If m, n are equal to the size of A, a faster method can be used for
    calculating the local sum.  Otherwise, the slower but more general method
    can be used.  The faster method is more than twice as fast and is also
    less memory intensive.
    
    As it is currently called (2021-04-12), the `else` case appears to never
    be invoked
    """
    if m == A.shape[0] and n == A.shape[1]:
        s = np.cumsum(A, axis=0)
        secondPart = np.tile(s[-1], (m-1, 1)) - s[0:-1]
        c = np.concatenate((s, secondPart), axis=0)
        s2 = np.cumsum(c, axis=1)
        lastColumn = s2[:, -1].reshape((s2.shape[0], 1))
        secondPart = np.tile(lastColumn, (1, n-1)) - s2[:, 0:-1]
        local_sum_A = np.concatenate((s2, secondPart), axis=1)
    else:
        # breal the padding into parts to save on memory
        B = np.zeros((A.shape[0] + 2*m, A.shape[1]))

#        B(m+1:m+size(A,1),:) = A
#        s = cumsum(B,1)
#        c = s(1+m:end-1,:)-s(1:end-m-1,:)
#        d = zeros(size(c,1),size(c,2)+2*n)
#        d(:,n+1:n+size(c,2)) = c
#        s = cumsum(d,2)
#        local_sum_A = s(:,1+n:end-1)-s(:,1:end-n-1)
        local_sum_A = 0

    return local_sum_A


# we assume that we only deal with float number
def _shift_data(A):
    """
    Convert array to type Float, and shift the data range to be greater than zero
    """
    B = A.astype(float)

    if not np.issubdtype(A.dtype, np.unsignedinteger):
        min_B = np.min(B)
        if min_B < 0:
            B -= min_B
    return B


def _find_closest_valid_dimension(n):

    # Find the closest valid dimension above the desired dimension.  This
    # will be a combination of 2s, 3s, and 5s.

    # Incrementally add 1 to the size until
    # we reach a size that can be properly factored.
    new_number = n
    result = 0
    new_number -= 1
    while not result == 1:
        new_number += 1
        result = _factorize_number(new_number)

    return new_number


def _factorize_number(n):
    for ifac in np.array([2, 3, 5]):
        while np.fmod(n, ifac) == 0:
            n = n / ifac
    return n



'''
Provies a function to generate egocentric maps from tracking/spike data
'''
import numpy as np

from opexebo import defaults as default



def egocentric_occupancy(positions, head_angles, num_angles, num_bins, boundaries, arena_shape):
    '''Generate a 2D histogram of (angle, border_distance) data . Can calculate either allocentric
    or egocentric values. To calculate allocentric (=fixed reference), set head_angles to None
    
    Implementation is based on, but indepdent of, https://doi.org/10.1016/j.cub.2019.07.007
    
    This is a moderately general function: it will work for either the tracking
    or spike data. However, it is sensible to enforce that the `boundaries` keyword
    uses the same value for both sides if the goal is to generate a meaningful ratemap
    
    Parameters
    ----------
    position: np.ndarray
        array of (x, y) positions of animal
    head_angle: np.ndarray -or- None
        Angle [RADIANS] of the animal's head with respect to the X axis
        Zero radians points along the x axis. 
        Set to None to calculate the allocentric values
    num_angles: int
        Number of angles at which to calculate the wall distance
    num_bins:
        Number of bins to consider wall distance
    boundaries: tuple
        Some way of defining where the boundary of the arena is
        if `arena_shape` is `square`, then boundaries should be:
            ( (min_x, max_x), (min_y, max_y) )
        elif `arena_shape` is `circ`, then boundaries should be:
            ( (centre_x, centre_y), radius)
    arena_shape: str
        definition of the shape of the arena, ties in to format of `boundaries`
        Acceptable values are given in `opexebo.defaults.shapes_*
    
    Returns
    -------
    hist: np.ndarray
        2D histogram of (wall_distance, angle) data
        hist[0, :] are all the distance bins for a single angle
        hist[:, 0] are all the angles for a single distance bin
    distance_bins: np.ndarray
        Distance bins, +1 element longer than hist[0, :].shape
    angle_bins: np.ndarray
        Angular bins, +1 element longer than hist[:, 0].shape   
    '''
    if positions.shape[0] == 2:
        # The easiest way to generate positions is `np.array((pos_x, pos_y))`, but that is the
        # transpose of what we want - so handle this automatically
        positions = positions.T
    
    # Create our output data structure
    num_positions = positions.shape[0]
    distances = np.zeros((num_positions, num_angles))
    
    # Create our target angle centres
    target_angles = (np.arange(num_angles) * 2 * np.pi / num_angles)
    target_angles += 0.5*np.diff(target_angles)
    
    if not isinstance(head_angles, np.ndarray):
        if isinstance(head_angles, (tuple, list)):
            head_angles = np.array(head_angles)
        elif head_angles is None:
        # Shortcut to let the user specify allocentric calculation
        # In effect, we do every calculation with a head_angle of 0
            head_angles = np.zeros(num_positions)
    
    # Calculate the wall distances
    for i, pos in enumerate(positions):
        distances[i] = _wall_distance(pos, head_angles[i], target_angles, boundaries, arena_shape)
    
    # Generating a histogram is a bit weird here, because our data is semi-histogrammed
    # already, i.e. the target angles have already been discretised. 
    # Thus, we call 1D histogram repeatedly, once for each target_angle
    
    # Calculate our histogram limits:
    # distance can be from 0 to the maximum diagonal length (diameter for a circle)
    min_bin = 0
    if arena_shape in default.shapes_square:
        max_bin = np.sqrt(np.square(boundaries[0][1] - boundaries[0][0]) + 
                          np.square(boundaries[1][1] - boundaries[1][0]))
    elif arena_shape in default.shapes_circle:
        max_bin = boundaries[1] * 2
    
    distance_bins = np.linspace(min_bin, max_bin, num_bins+1) # 1 longer because it lists the (start, end) of each bin, not just the centre
    angle_bins = np.linspace(0, 2*np.pi, num_angles+1)
    
    data = np.zeros((num_angles, num_bins))
    for i, tad in enumerate(distances.T):
        # we wrote this with each row being a single frame
        # but now we want to iterate over the tracking angles
        # hence transposing
        data[i, :], distance_bins = np.histogram(tad, distance_bins)
    return data, distance_bins, angle_bins
        
    
    
        


def _wall_distance(position, head_angle, target_angles, boundaries, arena_shape):
    '''For a single position and head angle, generate an array of border-distances
    for a given list of target angles. 
    
    Parameters
    ----------
    position: tuple
        (x, y) position of animal
    head_angle: float
        Angle [RADIANS] of the animal's head with respect to the X axis
        Zero radians points towards positive x values
        To calculate the allocentric values, set to zero
    target_angles: np.ndarray
        List of angles [RADIANS] at which to calculate the distance-to-wall
    boundaries: ???
        Some way of defining where the boundary of the arena is
        if `arena_shape` is `square`, then boundaries should be:
            ( (min_x, max_x), (min_y, max_y) )
        elif `arena_shape` is `circ`, then boundaries should be:
            ( (centre_x, centre_y), radius)
    arena_shape: str
        definition of the shape of the arena, ties in to format of `boundaries`
    
    Returns
    -------
    np.ndarray
        Distances to bounding wall at egocentric angles defined by `target_angles`
    '''    
    distances = np.zeros(target_angles.size)
    x, y = position
    target_angles = (target_angles + head_angle)%(2*np.pi)
    
    if arena_shape in default.shapes_square:
        '''Optimised array operation for square arenas
        Expects to recieve `boundaries` as a tuple of tuples
        ( ( min(x), max(x)), (min(y), max(y)) )
        Ca. 75x speedup over the for-loop version'''
        bounds_x, bounds_y = boundaries
        
        # Calculate the relevant boundary in x and y for each angle
        x_bounds = np.zeros(target_angles.shape)
        x_criteria = (np.abs(target_angles-np.pi)) < (np.pi*0.5)
        x_bounds[x_criteria] = bounds_x[0]
        x_bounds[np.logical_not(x_criteria)] = bounds_x[1]
        
        y_bounds = np.zeros(target_angles.shape)
        y_criteria = target_angles > np.pi
        y_bounds[y_criteria] = bounds_y[0]
        y_bounds[np.logical_not(y_criteria)] = bounds_y[1]
        
        # Calculate the distance to the vertical [0] and horizontal [1] boundaries
        temp_distances = np.zeros((2, target_angles.size))
        temp_distances[0] = (x_bounds - x) / np.cos(target_angles)
        temp_distances[1] = (y_bounds - y) / np.sin(target_angles)
        
        # Take the shortest value as the actual distance
        distances = np.min(temp_distances, axis=0)
        
    elif arena_shape in default.shapes_circle:
        raise NotImplementedError("Circular arenas are not currently supported")
    else:
        raise NotImplementedError(f"Arena Shape '{arena_shape}' are not currently supported")
    return distances


""" Tests for spatialCrossCorrelation"""
import numpy as np
import pytest

from opexebo.general import spatial_cross_correlation as func

print("=== tests_general_spatial_cross_correlation ===")


###############################################################################
################                HELPER FUNCTIONS
###############################################################################


###############################################################################
################                MAIN TESTS
###############################################################################


def test_invalid_inputs():
    with pytest.raises(ValueError):  # 1d array, mismatched sizes
        arr0 = np.zeros(5)
        arr1 = np.zeros(6)
        func(arr0, arr1)
    with pytest.raises(ValueError):  # 2d array, mismatched sizes
        arr0 = np.zeros((3, 3))
        arr1 = np.zeros((3, 4))
        func(arr0, arr1)
    with pytest.raises(ValueError):  # Random arguments
        arr0 = np.zeros(5)
        arr1 = "b"
        func(arr0, arr1)
    print("test_invalid_inputs passed")


def test_trivial_inputs_1d():
    num_bins = 10
    """Correlation of all-zeros should be NaN"""
    arr0 = np.zeros(num_bins)
    arr1 = np.zeros(num_bins)
    p = func(arr0, arr1)
    assert np.isnan(p[0])

    """Correlation of equal np.arange"""
    arr0 = np.arange(num_bins)
    arr1 = np.arange(num_bins)
    p = func(arr0, arr1)
    assert np.isclose(p[0], 1)

    """Correlation of equal np.arange"""
    arr0 = np.arange(num_bins)
    arr1 = np.flip(np.arange(num_bins))
    p = func(arr0, arr1)
    assert np.isclose(p[0], -1)

    """Handling occasionally NaNs"""
    arr0 = np.arange(num_bins).astype(float)
    arr1 = np.arange(num_bins).astype(float)
    arr0[2] = np.nan
    arr1[5] = np.nan
    p = func(arr0, arr1)
    assert np.isclose(p[0], 1)
    print("test_trivial_inputs_1d passed")


def test_trivial_inputs_2d():
    num_bins = (10, 11)
    """Correlation of all-zeros should be NaN"""
    arr0 = np.zeros(num_bins)
    arr1 = np.zeros(num_bins)
    p = func(arr0, arr1)
    assert np.isnan(p[0])
    assert np.isnan(p[1]).all()

    """correlation of equal arange"""
    arr0 = np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins)
    arr1 = np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins)
    p = func(arr0, arr1, row_major=True)
    assert np.isclose(p[0], 1)
    assert np.isclose(p[1], 1).all()
    p = func(arr0, arr1, row_major=False)
    assert np.isclose(p[1], 1).all()

    """correlation of opposite arange"""
    arr0 = np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins)
    arr1 = np.flip(np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins))
    p = func(arr0, arr1, row_major=True)
    assert np.isclose(p[0], -1)
    assert np.isclose(p[1], -1).all()
    p = func(arr0, arr1, row_major=False)
    assert np.isclose(p[1], -1).all()

    """Handling occasional NaNs"""
    arr0 = np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins).astype(float)
    arr1 = np.arange(num_bins[0] * num_bins[1]).reshape(*num_bins).astype(float)
    arr0[2, 3] = np.nan
    arr1[4, 8] = np.nan
    p = func(arr0, arr1, row_major=True)
    assert np.isclose(p[0], 1)
    assert np.isclose(p[1], 1).all()
    assert np.sum(np.isnan(p[1])) == 0
    p = func(arr0, arr1, row_major=False)
    assert np.isclose(p[1], 1).all()
    assert np.sum(np.isnan(p[1])) == 0
    print("test_trivial_inputs_2d passed")


# if __name__ == "__main__":
#    test_trivial_inputs_1d()
#    test_trivial_inputs_2d()



import numpy as np

def bin_width_to_bin_number(length, bin_width):
    '''This conversion is done in several separate places, and must be done the
    same in every case. Therefore, refactor into a quick helper function.
    
    In cases where the bin_width is not a perfect divisor of length, the actual
    bins will be slightly smaller
    
    Parameters
    ----------
    length: float or np.ndarray
        Length of a side to be divided into equally spaced bins
    bin_width: float
        Dimension of a square bin or pixel
    
    Returns
    -------
    num_bins: int or np.ndarray
        Same type as `length`. Integer number of bins.
    '''
    if type(length) in (list, tuple):
        length = np.array(length)
    num_bins = np.ceil(length / bin_width).astype(int)
    return num_bins


"""Tests for WalkFilter"""
import numpy as np
import pytest

from opexebo.general import walk_filter as func
from opexebo import errors as err

print("=== tests_general_walk_filter ===")

speeds = np.arange(5000)
speed_cutoff = 1257.3
min_out = 1258
remaining = 3742

def test_remove():
    out = func(speeds, speed_cutoff, fmt="remove")
    assert out.size == remaining
    assert np.min(out) == min_out

def test_mask():
    out = func(speeds, speed_cutoff, fmt="mask")
    assert type(out) == np.ma.MaskedArray
    assert np.sum(out.mask) == min_out
    assert np.min(out) == min_out

def test_filter_other_arrays():
    positions = np.linspace(-1000, 1000, speeds.size)
    fspeed, out = func(speeds, speed_cutoff, positions, fmt="remove")
    assert fspeed.shape == out.shape
    


""" Tests for upsampling"""
import numpy as np
import pytest

from opexebo.general import upsample as func

print("=== tests_upsampling ===")


###############################################################################
################                MAIN TESTS
###############################################################################


def test_invalid_inputs():
    # Invalid array format
    with pytest.raises(NotImplementedError):
        upscale = 2
        func(3, upscale)
        func("abc", upscale)
        func({2: 2}, upscale)

    # Invalid integerr upscaling
    # Force integer upsclaing with masked array inputs
    with pytest.raises(NotImplementedError):
        masked_array = np.ma.ones((4, 6))
        high_dim_ma = np.ma.ones((4, 6, 7))
        fractional_upscale = 2.5
        shrinking_upscale = 0
        valid_upscale = 2
        func(masked_array, fractional_upscale)
        func(masked_array, shrinking_upscale)
        func(high_dim_ma, valid_upscale)

    # Invalid fractional upscaling
    # force fractional upscaling with floating point upscale
    with pytest.raises(ValueError):
        ndarray = np.ones((4, 6))
        zero_upscale = float(0)
        neg_upscale = -0.5
        func(ndarray, zero_upscale)
        func(ndarray, neg_upscale)

    print("test_invalid_inputs passed")


def test_integer_scaling_2d():
    upscale = int(2)
    ndarray = np.random.rand(50, 50)
    maarray = ndarray.copy()
    maarray[25:28, 10:16] = np.nan
    maarray = np.ma.masked_invalid(maarray)

    # Test behavior on masked array
    new_array = func(maarray, upscale)
    assert np.array_equal(np.array(maarray.shape) * upscale, np.array(new_array.shape))
    assert np.sum(maarray.mask) * (upscale ** 2) == np.sum(new_array.mask)

    # Test behavior on ndarray
    new_array = func(ndarray, upscale)
    assert np.array_equal(np.array(ndarray.shape) * upscale, np.array(new_array.shape))
    print("test_integer_scaling_2d passed")


def test_fractional_scaling_2d():
    upscale = 1.45
    ndarray = np.random.rand(50, 75)
    new_array = func(ndarray, upscale)
    assert np.array_equal(np.round(np.array(ndarray.shape) * upscale), new_array.shape)
    print("test_fractional_scaling_2d passed")


# if __name__ == '__main__':
#    test_invalid_inputs()
#    test_integer_scaling_2d()
#    test_fractional_scaling_2d()



import numpy as np

from opexebo import errors

def walk_filter(speed:np.ndarray, speed_cutoff:float, *args, fmt="remove"):
    """
    It is common practice when studying a freely moving subject to exclude data
    from periods when the subject was stationary, or nearly stationary. This
    method is described as a "walk-filter" - a high-pass filter on subject speed.
    
    This function allows an arbitrary number of arrays to be filtered in parallel
    to the speed (or whatever other filtering criteria are used). Filters can be
    performed either by removing the unwanted elements, or by masking them and
    retunring a MaskedArray.
    
    Example
    -------
    Filter speed only
    >>> speed = np.arange(500)
    >>> cutoff = 200
    >>> speed = walk_filter(speed, cutoff, fmt="remove")
    >>> speed.size
    300
    
    Filter other arrays as well
    >>> speed = np.arange(500)
    >>> cutoff = 200
    >>> pos_x = np.linspace(-25, 73, speed.size)
    >>> pos_y = np.linspace(0, 98, speed.size)
    >>> speed, pos_x, pos_y = walk_filter(speed, cutoff, pos_x, pos_y, fmt="remove")
    >>> speed.size
    300
    >>> pos_x.size
    300
    

    Parameters
    ----------
    speed : np.ndarray
        Array of speeds for other data points
    speed_cutoff : float
        The cutoff, below which values in ``speed`` will be excluded.
    *args : np.ndarray, optional
        Any other arrays that should be filtered in parallel with speed
        Optional arguments here _must_ be np.ndarrays with size equal to that of
        ``speed``
    fmt : str, optional
        Either "remove" or "mask". Determines how the values are returned
        "remove" (default) - the invalid valaues are removed from the array
        "mask" - the original array is returned as a MaskedArray, with the invalid
        values masked out.
    

    Returns
    -------
    np.ndarray
        Filtered copy of ``speed``
    [np.ndarray]
        Arbitrary other filtered arrays, if any other arrays were provided as *args
    """
    
    if not isinstance(speed, np.ndarray):
        raise errors.ArgumentError("`speed` should be an ndarray, not ({})".format(type(speed)))
    
    if not isinstance(speed_cutoff, (float, int)):
        raise errors.ArgumentError("`speed_cutoff` should be a numeric value ({})".format(type(speed_cutoff)))
    if speed_cutoff <= 0 or not np.isfinite(speed_cutoff):
        raise errors.ArgumentError("\speed_cutoff` should be a finite positive value")
    
    if fmt.lower() not in ("remove", "mask"):
        raise errors.ArgumentError("`fmt` should be either 'remove' or 'mask'")
    
    if len(args):
        for i, arg in enumerate(args):
            if not isinstance(arg, np.ndarray):
                raise errors.ArgumentError(f"`arg {i} is not a Numpy array ({arg})")
            if not arg.shape == speed.shape:
                raise errors.ArgumentError(f"`arg {i} is a different size to `speed`")
            
    good = speed >= speed_cutoff
    
    if fmt.lower() == "mask":
        bad = np.logical_not(good)
        speed = np.ma.masked_where(bad, speed)
        out_args = [np.ma.masked_where(bad, arg) for arg in args]
    elif fmt.lower() == "remove":
        speed = speed[good]
        out_args = [arg[good] for arg in args]
        
    if out_args:
        out_args.insert(0, speed)
        return out_args
    else:
        return speed



"""Python neurology toolbox

Subpackages
-----------
analysis
	specialized neuroscience analysis functions (gridness score, place maps,
    e.t.c.)
general
    general signal processing function (smoothing, correlation, e.t.c.)


defaults
    default values for keyword analysis parameters
"""
from . import defaults
from . import errors
from . import analysis
from . import general

__author__ = """Simon Ball"""
__email__ = 'simon.ball@ntnu.no'
__version__ = '0.7.1'



"""Provide a thead-direction turning curve analysis"""

import numpy as np
import opexebo
import opexebo.defaults as default


def tuning_curve(angular_occupancy, spike_angles, **kwargs):
    """Analogous to a RateMap - i.e. mapping spike activity to spatial position
    map spike rate as a function of angle

    Parameters
    ----------
    angular_occupancy : np.ma.MaskedArray
        unsmoothed histogram of time spent at each angular range
        Nx1 array, covering the range [0, 2pi] radians
        Masked at angles of zero occupancy
    spike_angles : np.ndarray
        Mx1 array, where the m'th value is the angle of the animal (in radians)
        associated with the m'th spike
    kwargs
        bin_width : float
            width of histogram bin in DEGREES
            Must match that used in calculating angular_occupancy
            In the case of a non-exact divisor of 360 deg, the bin size will be 
            shrunk to yield an integer bin number. 


    Returns
    -------
    tuning_curve : np.ma.MaskedArray
        unsmoothed array of firing rate as a function of angle
        Nx1 array

    Notes
    --------
    BNT.+analyses.turningcurve

    Copyright (C) 2019 by Simon Ball

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    """

    occ_ndim = angular_occupancy.ndim
    spk_ndim = spike_angles.ndim
    if occ_ndim != 1:
        raise ValueError("angular_occupancy must be a 1D array. You provided a"\
                         " %d dimensional array" % occ_ndim)
    if spk_ndim != 1:
        raise ValueError("spike_angles must be a 1D array. You provided a %d"\
                         " dimensional array" % spk_ndim)
    if np.nanmax(spike_angles) > 2*np.pi:
        raise Warning("Angles higher than 2pi detected. Please check that your"\
                      " spike_angle array is in radians. If it is in degrees,"\
                      " you can convert with 'np.radians(array)'")

    bin_width = kwargs.get("bin_width", default.bin_angle) # in degrees
    num_bins = opexebo.general.bin_width_to_bin_number(360., bin_width) # This is for validation ONLY, the value num_bins here is not passed onwards
    if num_bins != angular_occupancy.size:
        raise ValueError("Keyword 'bin_width' must match the value used to"\
                         " generate angular_occupancy")
    #UNITS!!!
    # bin_width and arena_size need to be in the same units. 
    # As it happens, I hardcoded arena_size as 2pi -> convert bin_width to radians
    # limits and spike_angles need to be in the same units
    
    bin_width = np.radians(bin_width)

    spike_histogram, bin_edges = opexebo.general.accumulate_spatial(spike_angles,
                arena_size=2*np.pi, limits=(0, 2*np.pi), bin_width=bin_width)

    tuning_curve = spike_histogram / (angular_occupancy + np.spacing(1))

    return tuning_curve



import math
import numpy as np


"""' Test helpers"""


def rms(img):
    return np.sqrt(np.nanmean(np.square(img)))


def compare_dict(a, b, verbose=False):
    """Compare that the contents of two dictionaries exactly match"""
    # check that keys match
    ka = sorted(a.keys())
    kb = sorted(b.keys())
    mismatched_keys = []
    assert ka == kb, "Keys do not match"
    if verbose:
        print("Keys match")
    equal_dicts = True  # Assume dicts are equal, negate on finding one or more non-matching values
    for k in ka:
        if isinstance(a[k], np.ndarray):
            match = np.allclose(a[k], b[k])
        elif isinstance(a[k], float):
            match = math.isclose(a[k], b[k], rel_tol=1e-06)
        else:
            match = a[k] == b[k]
        if not match:
            equal_dicts = False
            mismatched_keys.append(k)
            if verbose:
                print(f"key {k} does not match: {a[k]:} : {b[k]}")
    return equal_dicts, mismatched_keys


###############################################################################
################                HELPER FUNCTIONS
###############################################################################
def generate_2d_map(shape, bin_width, **kwargs):
    """Create a fake 2D histogram for testing functions that process 2D images
    e.g. SpatialOccupancy, RateMap, PlaceField, GridScore etc

    Parameters
    ----------
    shape : str
        Shape - either "rect" or "circ"
    bin_width : float
        Bins must be assumed square
    kwargs
        x : float
            x-dimension of a rectangular array
        y : float
            y-dimension of a rectangular array
        d : float
            diameter of a circular array
        coverage : float
            Probability of any given location being visited by the animal
            Given in the range [0,1]
        fields : various
            Define how firing fields are generated. Options:
                Undefined - no fields
                str - "random" - [5-15] randomly generated firing fields
                list of dicts - predefined firing fields per dictionary specifications
                    For keywords, see `generate_field`
    Returns
    -------
    arena : numpy.ndarray
        2D array
    """
    if shape == "rect":
        array_size = np.array((kwargs.get("y", 80), kwargs.get("x", 80)))
        num_bins = np.ceil(array_size / bin_width).astype(int)
        bin_width = array_size / num_bins
        bin_centres = [
            np.linspace(
                -1 * (array_size[i] - bin_width[i]) / 2,
                (array_size[i] - bin_width[i]) / 2,
                num_bins[i],
            )
            for i in range(2)
        ]
        arena = np.zeros(num_bins)
    elif shape == "circ":
        diameter = kwargs.get("d", 100)
        array_size = np.array([diameter, diameter])
        num_bins = np.ceil(diameter / bin_width).astype(int)
        arena = np.zeros((num_bins, num_bins))

        bin_width = diameter / num_bins
        bin_centres = (
            np.linspace(
                -(diameter - bin_width) / 2, (diameter - bin_width) / 2, num_bins
            ),
            np.linspace(
                -(diameter - bin_width) / 2, (diameter - bin_width) / 2, num_bins
            ),
        )
        distance = generate_distance_map(bin_centres)
        outside = distance > (diameter / 2)
        arena = np.ma.masked_where(outside, arena)
    else:
        raise NotImplementedError(
            f"Shape '{shape}' is not supported. Only 'rect' or 'circ' are supported"
        )
    # Generate 1 or more firing fields
    fields = kwargs.get("fields", None)
    if fields is None:
        print("No fields argument detected")
        # do nothing at all
        pass
    elif fields == "random":
        # Randomly generate firing fields
        num_fields = np.random.randint(5, 15)
        print(f"Randomfields: {num_fields}")
        for i in range(num_fields):
            x, y = np.random.randn(2) * array_size * 0.1  # randn: [-1, 1], normal dist
            s_major, s_minor = (
                np.random.rand(2) * array_size * 0.4
            )  # rand: [0, 1], linear dist
            theta = np.random.rand(1) * 2 * np.pi
            amplitude = np.random.rand(1) * 20
            noise = np.random.rand(1) * 0.2
            field = generate_field(
                bin_centres, x, y, s_major, s_minor, theta, amplitude, noise
            )
            arena += field
    elif isinstance(fields, list):
        print(f"Predefined fields detected: {len(fields)}")
        # A list of dictionaries of pre-defined fields:
        for field_dictionary in fields:
            field = generate_field(bin_centres, **field_dictionary)
            arena += field
    else:
        raise NotImplementedError(f"Fields definition not understood: <{fields}>")
    # field = generate_field(bin_centres, -5, 10, 10, 5, 0, 1, 0.2)
    # arena = arena + field

    # Generate non-perfect coverage
    coverage = kwargs.get("coverage", 1)
    visiting_probability = np.random.rand(*arena.shape)
    arena = np.ma.masked_where(visiting_probability > coverage, arena)
    return arena


def generate_distance_map(bin_centres):
    """Generate a map of distances from the centre, given arrays of the Y, X bin centres

    Parameters
    ----------
    bin_centres : list of 1d np.ndarrays
        List (y, x) of the centre locations of the bin edges

    Returns
    -------
    distances : 2d np.ndarray
        2D Numpy array of size (len(y), len(x)) where the values are distances from the centre.
    """
    X, Y = np.meshgrid(*bin_centres)
    distance = np.sqrt(np.square(X) + np.square(Y))
    return distance


def generate_field(bin_centres, x, y, s_major, s_minor, theta, amplitude, noise):
    """Generate a noisy firing field based on a Gaussian distribution

    Parameters
    ----------
    bin_centres : list of 1d numpy.ndarray
        List (y, x) of the centre locations of the bin edges
    x : float
        x co-ord of centre of Gaussian
    y : float
        y co-ord of centre of Gaussian
    s_major : float
        Standard deviation of long axis
    s_minor : float
        Standard deviation of short axis
    theta : float
        Angle of long axis w.r.t x-axis in radians
    amplitude : float
        Peak firing rate of field
    noise : float
        Noisiness of field, relative to amplitude

    Returns
    -------
    field : numpy.ndarray
        2D array, matching the sizes of bin edges, containing a single firing
        field.

    """
    sigma_x = (np.cos(theta) * s_major) + (np.sin(theta) * s_minor)
    sigma_y = (np.sin(theta) * s_major) - (np.cos(theta) * s_minor)
    X, Y = np.meshgrid(*bin_centres)

    field = np.exp(
        -0.5
        * (
            (np.square(X - x) / (2 * sigma_x ** 2))
            + (np.square(Y - y) / (2 * sigma_y ** 2))
        )
    )
    if not noise == 0:
        noise = np.random.rand(*field.shape) * noise
        noise[field < 0.05] = 0
    else:
        noise = 1

    field = field * noise * amplitude
    return field


def generate_hexagonal_grid_fields_dict():
    """Example code to generate a field definition list for a hexagonal grid of
    cells. Mimic a perfect grid cell."""
    sma = 5
    smi = 5
    amp = 1
    fields = []
    offset = 40
    for y_step in np.arange(-3, 4, 1):
        y = y_step * offset * 0.866
        for x_step in np.arange(-3, 4, 1):
            x = (x_step * offset) + (y_step % 2 * 0.5 * offset)
            field = {
                "x": x,
                "y": y,
                "s_major": sma,
                "s_minor": smi,
                "theta": 0,
                "amplitude": amp,
                "noise": 0,
            }
            fields.append(field)
    return fields



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from os import path

with open('readme.md') as readme_file:
    readme = readme_file.read()

with open('history.md') as history_file:
    history = history_file.read()

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'requirements.txt')) as f:
    requirements = f.read().split()

with open(path.join(here, 'optional-requirements.txt')) as f:
    optional_requirements = f.read().split()

setup_requirements = [ ]

test_requirements = [ ]

setup(
    name='opexebo',
    description="Collection of python code in Kavli lab.",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    url='https://github.com/kavli-ntnu/opexebo',
    author="Simon Ball",
    author_email='simon.ball@ntnu.no',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    install_requires=requirements,
    extras_require={"full": optional_requirements},
    include_package_data=True,
    keywords='neuroscience kavli gridscore',
    packages=find_packages(include=['opexebo*']),
    version='0.7.1',
    zip_safe=False,
)



"""Tests for SpeedScore"""
import numpy as np
import pytest

from opexebo.analysis import speed_score as func
from opexebo.errors import ArgumentError, SpeedBandwidthError

print("=== tests_analysis_speedScore===")


def test_invalid_inputs():
    n = 1000
    # Wrong dimensions for inputs
    # Tracking times, speeds not the same size
    with pytest.raises(ArgumentError):
        spike_times = np.sort(np.random.rand(n)) * n
        tracking_times = np.arange(n)
        tracking_speeds = np.random.rand(n + 10) * 30
        func(spike_times, tracking_times, tracking_speeds)

    # tracking_speeds wrong dimensions
    with pytest.raises(ArgumentError):
        spike_times = np.sort(np.random.rand(n)) * n
        tracking_times = np.arange(n)
        tracking_speeds = np.random.rand(n, 2) * 30
        func(spike_times, tracking_times, tracking_speeds)

    # Invalid bandpass type
    with pytest.raises(NotImplementedError):
        spike_times = np.sort(np.random.rand(n)) * n
        tracking_times = np.arange(n)
        tracking_speeds = np.random.rand(n) * 30
        bandpass = "xyz"
        func(spike_times, tracking_times, tracking_speeds, bandpass=bandpass)

    # Impossibly high speed_bandwidth
    with pytest.raises(SpeedBandwidthError):
        spike_times = np.sort(np.random.rand(n)) * n
        tracking_times = np.arange(n)
        tracking_speeds = np.random.rand(n) * 30
        kwargs = {
            "bandpass": "adaptive",
            "lower_bound_speed": 2,
            "upper_bound_time": 10,
            "speed_bandwidth": 0.01,
        }
        func(spike_times, tracking_times, tracking_speeds, **kwargs)


def test_random_inputs():
    """Try randomised inputs and check we get the right _kind_ of outputs"""
    n = 10000
    spike_times = np.sort(np.random.rand(n)) * n
    tracking_times = np.arange(n)
    tracking_speeds = np.random.rand(n) * 30
    for kwargs in (
        {"bandpass": "none"},
        {"bandpass": "fixed", "lower_bound_speed": 2, "upper_bound_speed": 10},
        {
            "bandpass": "adaptive",
            "lower_bound_speed": 2,
            "upper_bound_time": 10,
            "speed_bandwidth": 2,
        },
    ):

        ss, thresholds = func(spike_times, tracking_times, tracking_speeds, **kwargs)
        for val in ss.values():
            assert np.isfinite(val)
            assert -1 <= val <= 1


# if __name__ == '__main__':
#    test_invalid_inputs()
#    test_random_inputs()



'''
Provides function to calculate the spatial occupancy of the arena
'''
import numpy as np
import opexebo
from opexebo import defaults as default
from opexebo import errors



def spatial_occupancy(time, position, arena_size, **kwargs):
    '''
    Generate an occpuancy map: how much time the animal spent in each location
    in the arena.

    NOTES: This assumes that the positions have already been aligned and curated
    to remove NaNs. This is based on the expectation that it will primarily be
    used within the DataJoint framework, where the curation takes place at a
    much earlier stage.

    Parameters
    ----------
    time: np.ndarray
        timestamps of position data
    position: np.ndarray (x, [y])
        1d or 2d array of positions at timestamps. If 2d, then row major, such
        that `position[0]` corresponds to all `x`; and `position[1]` to all `y`
    arena_size: float or tuple of floats
        Dimensions of arena (in cm)
            * For a linear track, length
            * For a circular arena, diameter
            * For a rectangular arena, length or (length, length)
    arena_shape: {"square", "rect", "circle", "line"}
        Rectangular and square are equivalent. Elliptical or n!=4 polygons
        not currently supported. Defaults to Rectangular
    bin_width: float
        Bin size in cm. Default 2.5cm. If bin_width is supplied, `limit` must
        also be supplied. One of `bin_width`, `bin_number`, `bin_edges` must be
        provided
    bin_number: int or tuple of int
        Number of bins. If provided as a tuple, then `(x_bins, y_bins)`. One
        of `bin_width`, `bin_number`, `bin_edges` must be provided
    bin_edges: array-like
        Edges of the bins. Provided either as `edges` or `(x_edges, y_edges)`.
        One of `bin_width`, `bin_number`, `bin_edges` must be provided
    limits: tuple or np.ndarray
        (x_min, x_max) or (x_min, x_max, y_min, y_max)
        Provide concrete limits to the range over which the histogram searches
        Any observations outside these limits are discarded
        If no limits are provided, then use np.nanmin(data), np.nanmax(data)
        to generate default limits.
        As is standard in python, acceptable values include the lower bound
        and exclude the upper bound
    debug: bool, optional
        If `true`, print out debugging information throughout the function.
        Default `False`

    Returns
    -------
    masked_map: np.ma.MaskedArray
        Unsmoothed map of time the animal spent in each bin.
        Bins which the animal never visited are masked (i.e. the mask value is
        `True` at these locations)
    coverage: float
        Fraction of the bins that the animal visited. In range [0, 1]
    bin_edges: ndarray or tuple of ndarray
        x, or (x, y), where x, y are 1d np.ndarrays
        Here x, y correspond to the output histogram
    '''
    # Check for correct shapes.
    # `positions` may be either 1D (x only) or 2D (x only, or x and y)
    # Theremfore, must have ndim == 1 or 2
    # ANd if ndim == 2, must have min(shape) == 1 or 2
    dimensionality = position.ndim
    num_dimensions_data = min(position.shape)
    if dimensionality not in (1, 2):
        raise errors.ArgumentError(
                "Positions array must be a 1D or 2D array, you have provided a"
                " {}d array".format(dimensionality)
                )
    if dimensionality == 2 and num_dimensions_data not in (1, 2):
        raise errors.ArgumentError(
                "Positions array must contain either 1D data (x only) or 2D data"
                " (x, y). You have provided {}d data".format(num_dimensions_data)
                )

    # Handle NaN positions by converting to a Masked Array
    position = np.ma.masked_invalid(position)
    if not position.any():
        raise ValueError("The positions are entirely invalid")

    debug = kwargs.get("debug", False)

    if debug:
        print("Number of time stamps: %d" % len(time))
        print("Maximum time stamp value: %.2f" % time[-1])
        print("Time stamp delta: %f" % np.min(np.diff(time)))

    
    if dimensionality == 1:
        pos = position
    elif dimensionality == 2:
        pos = np.array([position[i, :] for i in range(min(position.shape))])

    occupancy_map, bin_edges = opexebo.general.accumulate_spatial(pos, arena_size, **kwargs)
    if debug:
        print(f"Frames included in histogram: {np.sum(occupancy_map)}"\
              f" ({np.sum(occupancy_map)/len(time):3})")

    # So far, times are expressed in units of tracking frames
    # Convert to seconds:
    frame_duration = np.min(np.diff(time))
    occupancy_map_time = occupancy_map * frame_duration

    if debug:
        print(f"Time length included in histogram: {np.sum(occupancy_map_time):2}"\
              f"({np.sum(occupancy_map_time)/time[-1]:3})")

    masked_map = np.ma.masked_where(occupancy_map < 0.001, occupancy_map_time)

    # Calculate the fractional coverage based on the mask. The occupancy_map is
    # zero where the animal has not gone, and therefore non-zero where the animal
    # HAS gone. . Coverage is 1.0 when the animal has visited every location
    # Does not take account of a circular arena, where not all locations are
    # accessible

    shape = kwargs.get("arena_shape", default.shape)

    if shape.lower() in default.shapes_square:
        coverage = np.count_nonzero(occupancy_map) / occupancy_map.size
    elif shape.lower() in default.shapes_circle:
        if isinstance(arena_size, (float, int)):
            diameter = arena_size
        elif isinstance(arena_size, (tuple, list, np.ndarray)):
            diameter = arena_size[0]
        in_field, _, _ = opexebo.general.circular_mask(bin_edges, diameter)
        coverage = np.count_nonzero(occupancy_map) / (np.sum(in_field))
        coverage = min(1.0, coverage)
        # Due to the thresholding, coverage might be calculated to be  > 1
        # In this case, cut off to a maximum value of 1.
    elif shape.lower() in default.shapes_linear:
        coverage = np.count_nonzero(occupancy_map) / occupancy_map.size
    else:
        raise NotImplementedError(f"Arena shape '{shape}' not understood")

    return masked_map, coverage, bin_edges



"""
Provide function for 2D placefield detection.
"""

import numpy as np
from scipy import ndimage
from skimage import measure, morphology

import opexebo
import opexebo.defaults as default
import opexebo.errors as err


def place_field(firing_map, **kwargs):
    '''
    Locate place fields on a firing map.

    Identifies place fields in 2D firing map. Placefields are identified by
    using an adaptive threshold. The idea is that we start with a peak value as
    the threshold. Then we gradually decrease the threshold until the field
    area doesn't change any more or the area explodes (this means the threshold
    is too low).

    Parameters
    ----------
    firing_map: np.ndarray or np.ma.MaskedArray
        smoothed rate map.
        If supplied as an np.ndarray, it is assumed that the map takes values
        of np.nan at locations of zero occupancy. If supplied as an np.ma.MaskedArray,
        it is assumed that the map is masked at locations of zero occupancy
    
    Other Parameters
    ----------------
    min_bins: int
        Fields containing fewer than this many bins will be discarded. Default 9
    min_peak: float
        Fields with a peak firing rate lower than this absolute value will
        be discarded. Default 1 Hz
    min_mean: float
        Fields with a mean firing rate lower than this absolute value will
        be discarded. Default 0 Hz
    init_thresh: float
        Initial threshold to search for fields from. Must be in the range [0, 1].
        Default 0.96
    search_method: str
        Peak detection finding method. By default, use `skimage.morphology.local_maxima`
        Acceptable values are defined in `opexebo.defaults`. Not required if 
        peak_coords are provided
    peak_coords: array-like
        List of peak co-ordinates to consider instead of auto detection. [y, x].
        Default None

    Returns
    -------
    fields: list of dict
        coords: np.ndarray
            Coordinates of all bins in the firing field
        peak_coords: np.ndarray
            Coordinates peak firing rate [y,x]
        centroid_coords: np.ndarray
            Coordinates of centroid (decimal) [y,x]
        area: int
            Number of bins in firing field. [bins]
        bbox: tuple
            Coordinates of bounding box including the firing field
            (y_min, x_min, y_max, y_max)
        mean_rate: float
            mean firing rate [Hz]
        peak_rate: float
            peak firing rate [Hz]
        map: np.ndarray
            Binary map of arena. Cells inside firing field have value 1, all
            other cells have value 0
    fields_map : np.ndarray
        labelled integer image (i.e. background = 0, field1 = 1, field2 = 2, etc.)

    Raises
    ------
    ValueError
        Invalid input arguments
    NotImplementedError
        non-defined peack searching methods

    Notes
    --------
    BNT.+analyses.placefieldAdaptive

    https://se.mathworks.com/help/images/understanding-morphological-reconstruction.html

    Copyright (C) 2018 by Vadim Frolov, (C) 2019 by Simon Ball, Horst Obenhaus
    '''
    ##########################################################################
    #####                   Part 1: Handle inputs
    # Get keyword arguments
    min_bins = kwargs.get("min_bins", default.firing_field_min_bins)
    min_peak = kwargs.get("min_peak", default.firing_field_min_peak)
    min_mean = kwargs.get("min_mean", default.firing_field_min_mean)
    init_thresh = kwargs.get("init_thresh", default.initial_search_threshold)
    search_method = kwargs.get("search_method", default.search_method)
    peak_coords = kwargs.get("peak_coords", None)
    debug = kwargs.get("debug", False)

    if not 0 < init_thresh <= 1:
        raise err.ArgumentError("Keyword 'init_thresh' must be in the range [0, 1]."\
                         f" You provided {init_thresh}")
    try:
        search_method = search_method.lower()
    except AttributeError:
        raise err.ArgumentError("Keyword 'search_method' is expected to be a string"\
                         f" You provided a {type(search_method)} ({search_method})")
    if search_method not in default.all_methods:
        raise err.ArgumentError("Keyword 'search_method' must be left blank or given a"\
                         f" value from the following list: {default.all_methods}."\
                         f" You provided '{search_method}'.")

    global_peak = np.nanmax(firing_map)
    if np.isnan(global_peak) or global_peak == 0:
        if debug:
            print(f"Terminating due to invalid global peak: {global_peak}")
        return [], np.zeros_like(firing_map)
    elif global_peak < 1:
        # I have observed issues with firing maps which have low absolute values (maximum value << 1)
        # Firing maps that have clearly visible firing fields, but with low absolute peaks (above the specified min_peak)
        # are dropped by the peak_search stage. Therefore, scale the entire map, and specified minimums, up in order to have
        # values >> 1
        target_peak_factor = 4
        if debug:
            print(
                "Absolute peak value is low ({}). Scaling such that the absolute peak factor is {}".format(
                    global_peak,
                    target_peak_factor
                )
            )
        scaling = True
        scaling_factor = target_peak_factor / global_peak
    else:
        scaling = False
        if debug:
            print("Absolute peak factor is > 1, no scaling applied")

    # Construct a mask of bins that the animal never visited (never visited -> true)
    # This needs to account for multiple input formats.
    # The standard that I want to push is that firing_map is type MaskedArray
        # In this case, the cells that an animal never visited have firing_map.mask[cell]=True
        # while firing_map.data[cell] PROBABLY = 0
    # An alternative is the BNT standard, where firing_map is an ndarray
        # In this case, the cells never visited are firing_map[cell] = np.nan
    # In either case, we need to get out the following:
        # finite_firing_map is an ndarray (float) where unvisted cells have a
        # meaningfully finite value (e.g. zero, or min())
        # mask is an ndarray (bool) where unvisited cells are True, all other cells are False

    if isinstance(firing_map, np.ma.MaskedArray):
        occupancy_mask = firing_map.mask
        finite_firing_map = firing_map.data.copy()
        finite_firing_map[np.isnan(firing_map.data)] = 0

    else:
        occupancy_mask = np.zeros_like(firing_map).astype('bool')
        occupancy_mask[np.isnan(firing_map)] = True
        finite_firing_map = firing_map.copy()
        finite_firing_map[np.isnan(firing_map)] = 0

    if scaling:
        finite_firing_map *= scaling_factor
        min_peak *= scaling_factor
        min_mean *= scaling_factor
        global_peak *= scaling_factor


    structured_element = morphology.disk(1)
    image_eroded = morphology.erosion(finite_firing_map, structured_element)
    fmap = morphology.reconstruction(image_eroded, finite_firing_map)
    
    ##########################################################################
    #####                   Part 2: find local maxima
    # Based on the user-requested search method, find the co-ordinates of local maxima
    if peak_coords is None:
        if search_method == default.search_method:
            peak_coords = opexebo.general.peak_search(fmap, **kwargs)
        elif search_method == "sep":
            #fmap = finite_firing_map
            peak_coords = opexebo.general.peak_search(fmap, **kwargs)
        else:
            raise NotImplementedError("The search method you have requested (%s) is"\
                                      " not yet implemented" % search_method)

    # obtain value of found peaks
    found_peaks = finite_firing_map[peak_coords[:, 0], peak_coords[:, 1]]

    # leave only peaks that satisfy the threshold
    good_peaks = (found_peaks >= min_peak)
    peak_coords = peak_coords[good_peaks, :]

    if debug:
        print(
            "Peak search outcome: {} peaks found; {} remaining after minimum peak constraint".format(
                len(found_peaks),
                len(good_peaks),
            )
        )


    ##########################################################################
    #####    Part 3: from local maxima get fields by expanding around maxima
    max_value = np.max(fmap)
    # prevent peaks with small values from being detected
    # SWB - This causes problems where a local peak is next to a cell that the animal never went
    # As that risks the field becoming the entire null region
    # Therefore, adding 2nd criterion to avoid adding information where none was actually known.
    fmap[np.logical_and(fmap < min_peak, fmap > 0.01)] = max_value * 1.5

    # this can be confusing, but this variable is just an index for the vector
    # peak_linear_ind
    peaks_index = np.arange(len(peak_coords))
    fields_map = np.zeros(fmap.shape, dtype=int)
    field_id = 1
    for i, peak_rc in enumerate(peak_coords):
        # peak_rc == [row, col]

        # select all peaks except the current one
        other_fields = peak_coords[peaks_index != i]
        if other_fields.size > 0:
            other_fields_linear = np.ravel_multi_index(
                        multi_index=(other_fields[:, 0], other_fields[:, 1]),
                        dims=fmap.shape, order='F')
        else:
            other_fields_linear = []

        used_th = init_thresh
        res = _area_change(fmap, occupancy_mask, peak_rc, used_th,
                           used_th-0.02, other_fields_linear)
        initial_change = res['acceleration']
        area2 = res['area2']
        first_pixels = np.nan
        if np.isnan(initial_change):
            for j in np.linspace(used_th+0.01, 1., 4):
                # Thresholds get higher, area should tend downwards to 1
                # (i.e. only including the actual peak)
                res = _area_change(fmap, occupancy_mask, peak_rc, j, j-0.01, other_fields_linear)
                initial_change = res['acceleration']
                area1 = res['area1']
                area2 = res['area2']
                # initial_change is the change from area1 to area 2
                # area2>area1 -> initial_change > 1
                # area2<area1 -> initial_change < 1
                # area 2 is calculated with lower threshold - should usually be larger
                first_pixels = res['first_pixels']
                if not np.isnan(initial_change) and initial_change > 0:
                    # True is both area1 and area2 are valid
                    # Weird conditonal from Vadim - initial change will EITHER:
                    # be greater than zero (can't get a negative area to give negative % change)
                    # OR be NaN (which will always yield false when compared to a number)
                    used_th = j - 0.01
                    break

            if np.isnan(initial_change) and not np.isnan(area1):
                # For the final change
                pixels = np.unravel_index(first_pixels, fmap.shape, 'F')
                fmap[pixels] = max_value * 1.5
                fields_map[pixels] = field_id
                field_id = field_id + 1

            if np.isnan(initial_change):
                # failed to extract the field
                # Do nothing and continue for-loop
                pass

        pixel_list = _expand_field(fmap, occupancy_mask, peak_rc, initial_change, area2,
                                   other_fields_linear, used_th)
        if np.any(np.isnan(pixel_list)):
            _, pixel_list, _ = _area_for_threshold(fmap, occupancy_mask,
                                                   peak_rc, used_th+0.01,
                                                   other_fields_linear)
        if len(pixel_list) > 0:
            pixels = np.unravel_index(pixel_list, fmap.shape, 'F')
        else:
            pixels = []

        fmap[pixels] = max_value * 1.5
        fields_map[pixels] = field_id
        field_id = field_id + 1

    if debug:
        print(
            "Field expansion around peaks: {} fields remain".format(
                len(np.unique(fields_map)) - 1, # zero means outside all fields
            )
        )


    ##########################################################################
    #####     Part 4: Determine which, if any, fields meet filtering criteria
    regions = measure.regionprops(fields_map)

    fields = []
    fields_map = np.zeros(finite_firing_map.shape)  # void it as we can eliminate some fields

    for region in regions:
        field_map = finite_firing_map[region.coords[:, 0], region.coords[:, 1]]
        mean_rate = np.nanmean(field_map)
        num_bins = len(region.coords)

        peak_rate = np.nanmax(field_map)
        peak_relative_index = np.argmax(field_map)
        peak_coords = region.coords[peak_relative_index, :]

        if num_bins >= min_bins and mean_rate >= min_mean:
            field = {}
            field['coords'] = region.coords
            field['peak_coords'] = peak_coords
            field['area'] = region.area
            field['bbox'] = region.bbox
            field['centroid_coords'] = region.centroid
            field['mean_rate'] = mean_rate
            field['peak_rate'] = peak_rate
            mask = np.zeros(finite_firing_map.shape)
            mask[region.coords[:, 0], region.coords[:, 1]] = 1
            field['map'] = mask

            fields.append(field)

            fields_map[region.coords[:, 0], region.coords[:, 1]] = len(fields)
        elif debug:
            # Print out some information about *why* the field failed
            if num_bins < min_bins:
                print("Field size too small (%d)" % num_bins)
            if mean_rate < min_mean:
                print("Field mean rate too low (%.2f Hz)" % mean_rate)
        else:
            # Field too small and debugging information not needed
            # Do nothing
            pass
    #fields_map = np.ma.masked_where(occupancy_mask, fields_map)
    return (fields, fields_map)



#########################################################
################        Helper Functions
#########################################################


def _expand_field(image, occupancy_mask, peak_rc, initial_change,
                  initial_area, other_fields_linear, initial_th):
    '''
    Adaptive placefield detection:
        Start with a threshold around 80%, step down in ~0.02
        Measure field at threshold
        If the field is invalid, take the previous iteration
        If field size has decreased, keep count
        If field size has increased by less than 300% of initial_change AND the
          field has decreased in size fewer than 3 times in a row
            If the field size hasn't changed in 10 steps, return the current field size
        Else
    '''
    pixel_list = np.nan
    last_area = initial_area
    last_pixels = []
    num_not_changing = 0
    num_decrease = 0

    num_steps = int((initial_th - 0.2) / 0.02) + 1
    for threshold in np.linspace(initial_th, 0.2, num_steps):
        threshold = np.round(threshold, 2)

        area, pixels, is_bad = _area_for_threshold(image, occupancy_mask, peak_rc,
                                                   threshold, other_fields_linear)
        if np.isnan(area) or is_bad:
            pixel_list = last_pixels
            break
        current_change = area / last_area
        if current_change < 1:
            num_decrease = num_decrease + 1
        else:
            num_decrease = 0

        if np.floor(current_change / initial_change) <= 2 and num_decrease < 3:
            if current_change == 1:
                num_not_changing = num_not_changing + 1
            else:
                num_not_changing = 0

            if num_not_changing < 10:
                last_area = area
                last_pixels = pixels
                continue

        pixel_list = last_pixels
        break

    peak_linear = np.ravel_multi_index(multi_index=(peak_rc[0], peak_rc[1]),
                                       dims=image.shape, order='F')
    # last_pixels is a vector, peak_linear is a single value
    if np.any(last_pixels == peak_linear):
        # good field
        pixel_list = last_pixels

    return pixel_list


def _area_change(image, occupancy_mask, peak_rc, first, second, other_fields_linear):
    ''' Compare the change in field area based on two threshold values

    If either threshold results in an invalid field (based on criteria in
    _area_for_threshold() ), then return NaNs to signal this fact

    Otherwise return information about the results from the two thresholds, and
    the % change in area

    Params
    ------
    image : np.ndarray
        Image, i.e. rate map being considered
    occupancy_mask: np.ndarray
        binary mask. True where the animal spent zero time
    peak_rc : tuple
        [row, col] of local maxima
    first : float
        first threshold
    second : float
        second threshold
    other_fields_linear : np.ndarray
        All other local maxima except the ony under consideration
    '''
    results = {'acceleration': np.nan, 'area1': np.nan, 'area2': np.nan,
               'first_pixels': np.nan,
               'second_pixels': np.nan}

    area1, first_pixels, is_bad1 = _area_for_threshold(image, occupancy_mask, peak_rc, first,
                                                       other_fields_linear)
    if np.isnan(area1) or is_bad1:
        return results

    area2, second_pixels, is_bad2 = _area_for_threshold(image, occupancy_mask, peak_rc, second,
                                                        other_fields_linear)
    if np.isnan(area2) or is_bad2:
        return results

    acceleration = area2 / area1 #* 100
    results['acceleration'] = acceleration
    results['area1'] = area1
    results['area2'] = area2
    results['first_pixels'] = first_pixels
    results['second_pixels'] = second_pixels

    return results


def _area_for_threshold(image, occupancy_mask, peak_rc, threshold, other_fields_linear):
    '''Calculate the area of the field defined by the local maxima 'peak_rc' and
    the relative thresholding value 'threshold'

    In addition, determine:
        * is the field contiguous (good), or does it contain voids? (bad)
        * Does the field extend to include other local maxima?

    1 - Based on the threshold, find the binary map of the field
    2 - Determine if there are any voids
        If there are any voids, then return an area of np.nan
    3 - Determine the co-ordinates of all cells inside the field. The area is
        given by the number of cells
        If any included cell ALSO appears in the list of 'other_fields_linear'
        then another local maxima has been included. In that case, return is_bad=True

    returns
    -------
    area : int
        Number of cells in field OR np.nan if field contains holes
    area_linear_indicies : np.ndarray
        indicies of all cells within field if field is valid
    is_bad : bool
        True IF field includes a second local maxima or IF field contains holes
    '''
    area = np.nan
    # Field is bad if it contains any other peak
    is_bad = False

    peak_value = image[peak_rc[0], peak_rc[1]]
    threshold_value = peak_value * threshold
    mask = (image >= threshold_value)

    # Mask includes all pixels above the desired threshold, including other disconnected fields
    # use morphology.label to exclude disconnected fields
    # connectivity=1 means only consider vertical/horizontal connections
    labeled_img = morphology.label(mask, connectivity=1)

    # we need to leave only one label that corresponds to the peak
    target_label = labeled_img[peak_rc[0], peak_rc[1]]
    labeled_img[labeled_img != target_label] = 0
    labeled_img[labeled_img == target_label] = 1

    #labelled_img only includes cells that are:
    #   Above the current threshold
    #   Connected (V/H) to the currently considered local maxima

    # calclate euler_number by hand rather than by regionprops
    # This yields results that are more similar to Matlab's regionprops
    # NOTE - this uses scipy.ndimage.morphology, while most else uses skimage.morphology
    filled_image = ndimage.binary_fill_holes(labeled_img)
    euler_array = (filled_image != labeled_img)  # True where holes were filled in

    euler_array = np.maximum((euler_array*1) - (occupancy_mask*1), 0)
    # Ignore filled-in holes if it is due to the animal never visiting that location
    # Convert both arrays to integer, subtract one from the other, and replace resulting -1 values with 0
    # NOTE! np.maximum is element-wise, i.e. it returns an array. This is DIFFERENT to np.max, which returns a float.

    euler_objects = morphology.label(euler_array, connectivity=2) # connectivity=2 : vertical, horizontal, and diagonal
    num = np.max(euler_objects) # How many holes were filled in
    euler_number = -num + 1

    if euler_number <= 0:
        # If any holes existed, then return this
        is_bad = True
        return (area, [], is_bad)

    regions = measure.regionprops(labeled_img)
    area = np.sum(labeled_img == 1)
    area_linear_indices = np.ravel_multi_index(multi_index=(regions[0].coords[:, 0],
            regions[0].coords[:, 1]), dims=image.shape, order='F') # co-ordinates of members of field
    if len(other_fields_linear) > 0:
        is_bad = len(np.intersect1d(area_linear_indices, other_fields_linear)) > 0 # True if any other local maxima occur within this field

    return (area, area_linear_indices, is_bad)
